import pytest

from acceptance import assert_no_repetition, path_to_cnf_files, reset_expected_solutions
from sweetpea import *
from sweetpea._internal.encoding_diagram import print_encoding_diagram
from sweetpea._internal.server import build_cnf

color      = Factor("color",  ["red", "blue", "green"])
word       = Factor("motion", ["red", "blue"])

def illegal_stimulus(color, word):
    return color == "green" and word == "blue"

def legal_stimulus(color, word):
    return not illegal_stimulus(color, word)

stimulus_configuration = Factor("stimulus configuration", [
    DerivedLevel("legal",   WithinTrial(legal_stimulus, [color, word])),
    DerivedLevel("illegal", WithinTrial(illegal_stimulus, [color, word]))
])

constraints = [Exclude(stimulus_configuration["illegal"])]

design       = [color, word, stimulus_configuration]
crossing     = [color, word]


@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_no_solutions_without_override_flag(strategy):
    block       = CrossBlock(design, crossing, constraints)
    experiments = synthesize_trials(block, 500, sampling_strategy=strategy)

    assert block.crossing_size() == 5
    assert len(experiments) == 0
    assert_no_repetition(experiments)

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_override_flag(strategy):
    block       = CrossBlock(design, crossing, constraints, require_complete_crossing=False)
    experiments = synthesize_trials(block, 500, sampling_strategy=strategy)

    assert block.crossing_size() == 5
    assert len(experiments) == 120
    assert_no_repetition(experiments)  # FIXME

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_exclusion_via_complex_factor(strategy):
    def illegal_stimulus(color, word):
        return color[0] == "green" and word[0] == "blue"
    def legal_stimulus(color, word):
        return not illegal_stimulus(color, word)

    stimulus_configuration = Factor("stimulus configuration", [
        DerivedLevel("legal",   Transition(legal_stimulus, [color, word])),
        DerivedLevel("illegal", Transition(illegal_stimulus, [color, word]))
    ])

    constraints = [Exclude(stimulus_configuration["illegal"])]

    design       = [color, word, stimulus_configuration]
    crossing     = [color, word]

    block      = CrossBlock(design, crossing, constraints,
                            require_complete_crossing=False)
    experiments = synthesize_trials(block=block, samples=150, sampling_strategy=strategy)

    assert len(experiments) == 120

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_exclusion_via_nested_complex_factor(strategy):
    def unhappy_stimulus(color, word):
        return color[0] == "green" and word[0] == "blue"
    def happy_stimulus(color, word):
        return not unhappy_stimulus(color, word)
    smiley = Factor("smiley", [
        DerivedLevel("happy",   Transition(happy_stimulus, [color, word])),
        DerivedLevel("unhappy", Transition(unhappy_stimulus, [color, word]))
    ])

    def illegal_stimulus(smiley):
        return smiley == "unhappy"
    def legal_stimulus(smiley):
        return smiley == "happy"
    stimulus_configuration = Factor("stimulus configuration", [
        DerivedLevel("legal",   WithinTrial(legal_stimulus, [smiley])),
        DerivedLevel("illegal", WithinTrial(illegal_stimulus, [smiley]))
    ])

    constraints = [Exclude(stimulus_configuration["illegal"])]

    design       = [color, word, smiley, stimulus_configuration]
    crossing     = [color, word]

    block      = CrossBlock(design, crossing, constraints,
                            require_complete_crossing=False)
    experiments = synthesize_trials(block=block, samples=150, sampling_strategy=strategy)

    assert len(experiments) == 120

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_override_flag_and_multiple_trials_excluded(strategy):
    # With this constraint, there should only be ONE allowed crossing, and therefore one solution.
    constraints = [Exclude(stimulus_configuration["legal"])]
    block       = CrossBlock(design, crossing, constraints, require_complete_crossing=False)
    experiments = synthesize_trials(block, 500, sampling_strategy=strategy)

    assert block.crossing_size() == 1
    assert len(experiments) == 1
    assert_no_repetition(experiments)

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_crossing_levels_excluded(strategy):
    color = Factor("color", ['red', 'green'])
    size = Factor("size", ['small', 'med', 'large'])

    red = color.get_level('red')
    med = size.get_level('med')

    match = Factor(name="match", initial_levels=[
        DerivedLevel(name="up", window=WithinTrial(predicate=lambda a, b: b != 'large', factors=[color, size])),
        DerivedLevel(name="down", window=WithinTrial(predicate=lambda a, b: b == 'large', factors=[color, size]))
    ])

    block      = CrossBlock(design=[color, size, match],
                            crossing=[color, match],
                            constraints=[Exclude(red)],
                            require_complete_crossing=False)
    experiments = synthesize_trials(block=block, samples=200, sampling_strategy=strategy)

    assert len(experiments) == 4
    assert len(experiments[0][list(experiments[0].keys())[0]]) == 2

    for e in experiments:
        for l in e['color']:
            assert l != 'red'

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_count_with_design_levels_excluded(strategy):
    color = Factor("color", ['red', 'green'])
    size = Factor("size", ['small', 'med', 'large'])
    
    red = color.get_level('red')
    med = size.get_level('med')
    
    match = Factor(name="match", initial_levels=[
        DerivedLevel(name="up", window=WithinTrial(predicate=lambda a, b: b != 'large', factors=[color, size])),
        DerivedLevel(name="down", window=WithinTrial(predicate=lambda a, b: b == 'large', factors=[color, size]))
    ])
    down = match.get_level('down')

    block      = CrossBlock(design=[color, size, match], crossing=[color, size],
                            constraints=[Exclude(down)],
                            require_complete_crossing=False)
    experiments = synthesize_trials(block=block, samples=200, sampling_strategy=strategy)

    assert len(experiments) == 24
    assert len(experiments[0][list(experiments[0].keys())[0]]) == 4
    
    for e in experiments:
        for l in e['size']:
            assert l != 'large'

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_with_extra_crossing_factor(strategy):
    color      = Factor("color",  ["red", "green", "blue", "brown"])
    word       = Factor("word", ["Red", "Green", "Blue", "Brown"])
    location       = Factor("location", ["up", "down", "left", "right"])

    def is_response_left(color):
        return color == "red"
    def is_response_right(color):
        return color == "green"
    def is_response_up(color):
        return color == "blue"
    def is_response_down(color):
        return color == "brown"

    response = Factor("response", [
        DerivedLevel("left", WithinTrial(is_response_left, [color])),
        DerivedLevel("right", WithinTrial(is_response_right, [color])),
        DerivedLevel("up", WithinTrial(is_response_up, [color])),
        DerivedLevel("down", WithinTrial(is_response_down, [color]))
    ])

    def is_congruent(color, word):
        return color[1:] == word[1:]

    def is_incongruent(color, word):
        return not is_congruent(color, word)

    congruent = DerivedLevel("congruent", WithinTrial(is_congruent, [color, word]))
    incongruent = DerivedLevel("incongruent", WithinTrial(is_incongruent, [color, word]))

    congruency = Factor("congruency", [
        congruent,
        incongruent
    ])


    constraints = [Exclude(congruent)]

    design       = [color, word, response, congruency, location]
    crossing     = [color, word, location]
    block        = CrossBlock(design, crossing, constraints,
                              require_complete_crossing=False)

    experiments  = synthesize_trials(block, 1, sampling_strategy=strategy)

    assert len(experiments[0]['color']) == 48 # would be 64 without exclude

    found = {}
    for i in range(len(experiments[0]['color'])):
        key = (experiments[0]['color'][i], experiments[0]['word'][i], experiments[0]['location'][i])
        assert not key in found
        found[key] = True

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solution_with_just_one_factor(strategy):
    color = Factor("color", ['red', 'green'])
    red = color.get_level('red')

    block      = CrossBlock(design=[color], crossing=[color],
                            require_complete_crossing=False,
                            constraints=[Exclude(red)])
    experiments = synthesize_trials(block=block, samples=1, sampling_strategy=strategy)

    assert len(experiments[0]['color']) == 1
    assert experiments[0]['color'][0] == 'green'

def test_correct_solution_count_with_override_flag_and_multiple_trials_excluded_cnf():
    # With this constraint, there should only be ONE allowed crossing, and therefore one solution.
    constraints = [Exclude(stimulus_configuration["legal"])]
    block       = CrossBlock(design, crossing, constraints, require_complete_crossing=False)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_with_override_flag_and_multiple_trials_excluded.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_with_override_flag_and_multiple_trials_excluded.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()  # FIXME

@pytest.mark.parametrize('strategy', [RandomGen, IterateSATGen])
def test_correct_solutions_with_implicitly_excluded_crossing_due_to_derived_definition(strategy):
    def ugly_stimulus(color, word):
        return color == "green" and word == "red"
    def pretty_stimulus(color, word):
        return not ugly_stimulus(color, word)
    aesthetic = Factor("aesthetic", [
        DerivedLevel("ugly",   WithinTrial(ugly_stimulus, [color, word])),
        DerivedLevel("pretty", WithinTrial(pretty_stimulus, [color, word]))
    ])


    def unhappy_stimulus(color, word):
        return color[0] == "green" and word[0] == "blue"
    def happy_stimulus(color, word):
        return not unhappy_stimulus(color, word)
    smiley = Factor("smiley", [
        DerivedLevel("happy",   Transition(happy_stimulus, [color, word])),
        DerivedLevel("unhappy", Transition(unhappy_stimulus, [color, word]))
    ])

    constraints = [Exclude(aesthetic["ugly"])]

    design       = [color, word, smiley, aesthetic]
    crossing     = [color, word, aesthetic]


    block       = CrossBlock(design, crossing, constraints, require_complete_crossing=False)
    experiments = synthesize_trials(block, 4, sampling_strategy=RandomGen)

    assert block.crossing_size() == 5
    assert len(experiments) == 4
    for e in experiments:
        assert len(e['color']) == 5
        seen = set()
        for i in range(5):
            key = (e['color'][i], e['motion'][i])
            assert not key in seen
            seen.add(key)



import operator as op
import pytest
from typing import cast

from itertools import permutations

from sweetpea import CrossBlock
from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window
from sweetpea._internal.constraint import Exclude


color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])
size  = Factor("size",  ["big", "small", "tiny"])
direction = Factor("direction", ["up", "down"])

red_color = color["red"]
red_text = text["red"]
blue_color = color["blue"]
blue_text = text["blue"]

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

yes_color_repeats = color_repeats_factor["yes"]
no_color_repeats = color_repeats_factor["no"]

text_repeats_factor = Factor("repeated text?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])

yes_text_repeats = text_repeats_factor["yes"]
no_text_repeats = text_repeats_factor["no"]

congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])

yes_congruent = congruent_bookend["yes"]
no_congruent = congruent_bookend["no"]

color3 = Factor("color3", ["red", "blue", "green"])

yes_fn = lambda colors: colors[0] == colors[-1] == colors[-2]
no_fn = lambda colors: not yes_fn(colors)
color3_repeats_factor = Factor("color3 repeats?", [
    DerivedLevel("yes", Window(yes_fn, [color3], 3, 1)),
    DerivedLevel("no",  Window(no_fn, [color3], 3, 1))
])

yes_color3_repeats = color3_repeats_factor["yes"]
no_color3_repeats = color3_repeats_factor["no"]

def test_has_factor():
    block = CrossBlock([color, text], [color, text], [])

    assert block.has_factor(color) == color
    assert block.has_factor(Factor("bogus", ["red"])) == cast(Factor, None)


@pytest.mark.parametrize('design,expected',
    [([color, text, color_repeats_factor, text_repeats_factor], [0, 2, 16, 22]),
     ([color, text, text_repeats_factor, color_repeats_factor], [0, 2, 22, 16]),
     ([color_repeats_factor, text, color, text_repeats_factor], [2, 0, 16, 22]),
     ([text_repeats_factor, text, color, color_repeats_factor], [2, 0, 22, 16])])
def test_fully_cross_block_first_variable_for_factor(design, expected):
    block = CrossBlock(design,
                       [color, text],
                       # So {color,text}_repeats_factor is not treated as implied:
                       [Exclude(yes_color_repeats),
                        Exclude(yes_text_repeats)])

    assert block.first_variable_for_level(color, red_color) == expected[0]
    assert block.first_variable_for_level(color, blue_color) == expected[0] + 1
    assert block.first_variable_for_level(text, red_text) == expected[1]
    assert block.first_variable_for_level(text, blue_text) == expected[1] + 1
    assert block.first_variable_for_level(color_repeats_factor, yes_color_repeats) == expected[2]
    assert block.first_variable_for_level(color_repeats_factor, no_color_repeats) == expected[2] + 1
    assert block.first_variable_for_level(text_repeats_factor, yes_text_repeats) == expected[3]
    assert block.first_variable_for_level(text_repeats_factor, no_text_repeats) == expected[3] + 1


def test_fully_cross_block_first_variable_for_factor_with_color3():
    block = CrossBlock([color3_repeats_factor, color3, text],
                       [color3, text],
                       # So color3_repeats_factor is not treated as implied:
                       [Exclude(yes_color3_repeats)])

    assert block.first_variable_for_level(color3_repeats_factor, yes_color3_repeats) == 30
    assert block.first_variable_for_level(color3_repeats_factor, no_color3_repeats) == 31


def test_factor_variables_for_trial():
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       # So color_repeats_factor is not treated as implied:
                       [Exclude(yes_color_repeats)])

    assert block.factor_variables_for_trial(color, 1) == [1, 2]
    assert block.factor_variables_for_trial(color, 4) == [13, 14]

    assert block.factor_variables_for_trial(text, 2) == [7, 8]

    assert block.factor_variables_for_trial(color_repeats_factor, 2) == [18] # 17 is excluded
    assert block.factor_variables_for_trial(color_repeats_factor, 4) == [22] # 21 is excluded


def test_factor_variables_for_trial_with_expanded_crossing():
    # Because a transition is included in the crossing, this design requires 5 trials.
    block = CrossBlock([color, text, color_repeats_factor], [text, color_repeats_factor], [])

    assert block.factor_variables_for_trial(color, 1) == [1, 2]
    assert block.factor_variables_for_trial(color, 5) == [17, 18]

    assert block.factor_variables_for_trial(text, 2) == [7, 8]
    assert block.factor_variables_for_trial(text, 5) == [19, 20]

    assert block.factor_variables_for_trial(color_repeats_factor, 2) == [21, 22]
    assert block.factor_variables_for_trial(color_repeats_factor, 4) == [25, 26]
    assert block.factor_variables_for_trial(color_repeats_factor, 5) == [27, 28]


def test_variable_list_for_trial():

    # ---------------------------------------------------
    # |   Trial |  color   |   text   | repeated color? |
    # |       # | red blue | red blue |   yes      no   |
    # ---------------------------------------------------
    # |       1 |  1   2   |  3   4   |                 |
    # |       2 |  5   6   |  7   8   |    17      18   |
    # |       3 |  9   10  | 11   12  |    19      20   |
    # |       4 | 13   14  | 15   16  |    21      22   |
    # ---------------------------------------------------
    block = CrossBlock([color, text, color_repeats_factor],
                       [color, text],
                       # So color_repeats_factor is not treated as implied:
                       [Exclude(yes_color_repeats)])

    assert block.variable_list_for_trial(1) == [[ 1, 2 ], [ 3, 4 ], []]
    assert block.variable_list_for_trial(2) == [[ 5, 6 ], [ 7, 8 ], [18]]
    assert block.variable_list_for_trial(3) == [[ 9, 10], [11, 12], [20]]
    assert block.variable_list_for_trial(4) == [[13, 14], [15, 16], [22]]


def test_block_get_variable():
    block = CrossBlock([color, text, color_repeats_factor], [color, text], [])

    assert block.get_variable(1, (color, red_color)) == 1
    assert block.get_variable(1, (color, blue_color)) == 2
    assert block.get_variable(3, (color, red_color)) == 9
    assert block.get_variable(3, (color, blue_color)) == 10

    assert block.get_variable(2, (text, red_text)) == 7
    assert block.get_variable(2, (text, blue_text)) == 8
    assert block.get_variable(3, (text, red_text)) == 11
    assert block.get_variable(3, (text, blue_text)) == 12


def test_fully_cross_block_decode_variable():
    block = CrossBlock([color, text, color_repeats_factor, text_repeats_factor],
                       [color, text],
                       # So {color,text}_repeats_factor is not treated as implied:
                       [Exclude(yes_color_repeats),
                        Exclude(yes_text_repeats)])

    assert block.decode_variable(1) == (color, red_color)
    assert block.decode_variable(2) == (color, blue_color)
    assert block.decode_variable(5) == (color, red_color)
    assert block.decode_variable(14) == (color, blue_color)

    assert block.decode_variable(3) == (text, red_text)
    assert block.decode_variable(4) == (text, blue_text)
    assert block.decode_variable(15) == (text, red_text)
    assert block.decode_variable(12) == (text, blue_text)

    # assert block.decode_variable(17) == (color_repeats_factor, yes_color_repeats)
    assert block.decode_variable(18) == (color_repeats_factor, no_color_repeats)
    # assert block.decode_variable(19) == (color_repeats_factor, yes_color_repeats)
    assert block.decode_variable(22) == (color_repeats_factor, no_color_repeats)

    # assert block.decode_variable(23) == (text_repeats_factor, yes_text_repeats)
    assert block.decode_variable(24) == (text_repeats_factor, no_text_repeats)
    # assert block.decode_variable(27) == (text_repeats_factor, yes_text_repeats)
    assert block.decode_variable(28) == (text_repeats_factor, no_text_repeats)


def test_fully_cross_block_decode_variable_with_transition_first():
    block = CrossBlock([text_repeats_factor, text, color, color_repeats_factor],
                       [color, text],
                       # So {color,text}_repeats_factor is not treated as implied:
                       [Exclude(yes_color_repeats),
                        Exclude(yes_text_repeats)])

    assert block.decode_variable(1) == (text, red_text)
    assert block.decode_variable(2) == (text, blue_text)
    assert block.decode_variable(5) == (text, red_text)
    assert block.decode_variable(14) == (text, blue_text)

    assert block.decode_variable(3) == (color, red_color)
    assert block.decode_variable(4) == (color, blue_color)
    assert block.decode_variable(15) == (color, red_color)
    assert block.decode_variable(12) == (color, blue_color)

    # assert block.decode_variable(17) == (text_repeats_factor, yes_text_repeats)
    assert block.decode_variable(18) == (text_repeats_factor, no_text_repeats)
    # assert block.decode_variable(19) == (text_repeats_factor, yes_text_repeats)
    assert block.decode_variable(22) == (text_repeats_factor, no_text_repeats)

    # assert block.decode_variable(23) == (color_repeats_factor, yes_color_repeats)
    assert block.decode_variable(24) == (color_repeats_factor, no_color_repeats)
    # assert block.decode_variable(27) == (color_repeats_factor, yes_color_repeats)
    assert block.decode_variable(28) == (color_repeats_factor, no_color_repeats)


def test_fully_cross_block_decode_variable_with_general_window():
    block = CrossBlock([color, text, congruent_bookend],
                       [color, text],
                       [Exclude(no_congruent)])

    assert block.decode_variable(1) == (color, red_color)
    assert block.decode_variable(2) == (color, blue_color)
    assert block.decode_variable(5) == (color, red_color)
    assert block.decode_variable(14) == (color, blue_color)

    assert block.decode_variable(3) == (text, red_text)
    assert block.decode_variable(4) == (text, blue_text)
    assert block.decode_variable(15) == (text, red_text)
    assert block.decode_variable(12) == (text, blue_text)

    assert block.decode_variable(17) == (congruent_bookend, yes_congruent)
    # assert block.decode_variable(18) == (congruent_bookend, no_congruent)
    assert block.decode_variable(19) == (congruent_bookend, yes_congruent)
    # assert block.decode_variable(20) == (congruent_bookend, no_congruent)


def test_fully_cross_block_trials_per_sample():
    text_single  = Factor("text",  ["red"])

    assert CrossBlock([color, text],
                      [color, text],
                      []).trials_per_sample() == 4
    assert CrossBlock([color, text, direction],
                      [color, text, direction],
                      []).trials_per_sample() == 8
    assert CrossBlock([size, text_single],
                      [size, text_single],
                      []).trials_per_sample() == 3
    assert CrossBlock([size, color],
                      [size, color],
                      []).trials_per_sample() == 6
    assert CrossBlock([text_single],
                      [text_single],
                      []).trials_per_sample() == 1

    assert CrossBlock([color, text, color_repeats_factor], [color, text], []).trials_per_sample() == 4


def test_fully_cross_block_trials_per_sample_with_transition_in_crossing():
    block = CrossBlock([color, text, color_repeats_factor],
                       [text, color_repeats_factor],
                       [])

    # Typically, only 4 trials are needed to cross two factors each with two levels. (2 * 2 = 4)
    # However, because one of these factors is a transition, it doesn't apply to the first trial.
    # As a result, we actually need 5 trials to do a full crossing between the two.
    assert block.trials_per_sample() == 5

    # The crossing size is still just 4.
    assert block.crossing_size() == 4

def test_fully_cross_block_variables_per_trial():
    assert CrossBlock([color, text], [], []).variables_per_trial() == 4
    assert CrossBlock([color, text, con_factor], [], []).variables_per_trial() == 4
    assert CrossBlock([color, text, con_factor], [con_factor], []).variables_per_trial() == 6

    # Should exclude Transition and Windows from variables per trial count, as they don't always
    # have a representation in the first few trials. (Depending on the window width)
    assert CrossBlock([color, text, color_repeats_factor],
                      [color, text],
                      []).variables_per_trial() == 4


def test_fully_cross_block_grid_variables():
    assert CrossBlock([color, text, con_factor],
                      [color, text], []).grid_variables() == 16
    assert CrossBlock([color, text, con_factor],
                      [color, text],
                      [Exclude(con_level)]).grid_variables() == 12

    # Should include grid variables, as well as additional variables for complex windows.
    assert CrossBlock([color, text, color_repeats_factor],
                      [color, text],
                      [Exclude(yes_color_repeats)]).grid_variables() == 16


def test_fully_cross_block_variables_per_sample():
    assert CrossBlock([color, text, con_factor],
                      [color, text], []).variables_per_sample() == 16
    assert CrossBlock([color, text, con_factor],
                      [color, text],
                      [Exclude(con_level)]).variables_per_sample() == 12

    # Should include grid variables, as well as additional variables for complex windows.
    assert CrossBlock([color, text, color_repeats_factor],
                      [color, text],
                      [Exclude(yes_color_repeats)]).variables_per_sample() == 22

    assert CrossBlock([color, text, color_repeats_factor, text_repeats_factor],
                      [color, text],
                      [Exclude(yes_color_repeats),
                       Exclude(yes_text_repeats)]).variables_per_sample() == 28


def test_fully_cross_block_variables_for_factor():
    assert CrossBlock([color, text], [color, text], []).variables_for_factor(color) == 8
    assert CrossBlock([color, text], [color, text], []).variables_for_factor(text) == 8

    assert CrossBlock([color, text, color_repeats_factor],
                      [color, text],
                      []).variables_for_factor(color_repeats_factor) == 6
    assert CrossBlock([color, text, color_repeats_factor],
                      [color, text],
                      []).variables_for_factor(color_repeats_factor) == 6

    assert CrossBlock([color3_repeats_factor, color3, text],
                      [color3, text],
                      []).variables_for_factor(color3) == 18

    assert CrossBlock([color3_repeats_factor, color3, text],
                      [color3, text],
                      []).variables_for_factor(text) == 12

    assert CrossBlock([color3_repeats_factor, color3, text],
                      [color3, text],
                      []).variables_for_factor(color3_repeats_factor) == 8

    assert CrossBlock([color, text, congruent_bookend],
                      [color, text],
                      []).variables_for_factor(congruent_bookend) == 4


def test_fully_cross_block_crossing_size_with_exclude():
    # No congruent excludes 2 trials, 4 - 2 = 2
    assert CrossBlock([color, text, con_factor],
                      [color, text],
                      [Exclude(con_level)],
                      require_complete_crossing=False).crossing_size() == 2


def test_fully_cross_block_crossing_size_with_overlapping_exclude():
    # How about with two overlapping exclude constraints? Initial crossing size
    # should be 3 x 3 = 9.
    # Excluding congruent pairs will reduce that to 9 - 3 = 6
    # Then excluding red and green on top of that should make it 5.
    color = Factor("color", ["red", "blue", "green"])
    text  = Factor("text",  ["red", "blue", "green"])

    congruent_factor = Factor("congruent?", [
        DerivedLevel("congruent", WithinTrial(op.eq, [color, text])),
        DerivedLevel("incongruent", WithinTrial(op.ne, [color, text])),
    ])

    def illegal(color, text):
        return (color == "red" and text == "green") or color == text

    def legal(color, text):
        return not illegal(color, text)

    legal_factor = Factor("legal", [
        DerivedLevel("yes", WithinTrial(legal, [color, text])),
        DerivedLevel("no",  WithinTrial(illegal, [color, text]))
    ])

    assert CrossBlock([color, text, congruent_factor, legal_factor],
                      [color, text],
                      [Exclude(congruent_factor["congruent"]), # Excludes 3
                       Exclude(legal_factor["no"])], # Excludes 4, but 3 were already excluded
                      require_complete_crossing=False).crossing_size() == 5


def test_fully_cross_block_should_copy_input_lists():
    # CrossBlock should copy the input lists, so as not to break if the
    # user modifies the original list.
    design = [color, text, con_factor]
    crossing = [color, text]
    constraints = [Exclude(con_factor["con"])]

    block = CrossBlock(design, crossing, constraints)

    design.clear()
    assert len(block.design) == 3
    assert len(block.act_design) == 3

    crossing.clear()
    assert len(block.crossings[0]) == 2

    constraints.clear()
    assert len(block.constraints) == 5 # expanded constraints


def test_build_variable_list_for_simple_factors():
    block = CrossBlock([color, text, con_factor], [color, text], [Exclude(inc_level)])

    assert block.build_variable_lists((color, red_color)) == [[1, 7]]
    assert block.build_variable_lists((con_factor, con_factor["con"])) == [[5, 11]]


def test_build_variable_list_for_complex_factors():
    block = CrossBlock([color, text, color_repeats_factor], [color, text], [Exclude(no_color_repeats)])

    assert block.build_variable_lists((color_repeats_factor, yes_color_repeats)) == [[17, 19, 21]]
    assert block.build_variable_lists((color_repeats_factor, no_color_repeats))  == [[18, 20, 22]]


def test_build_variable_list_for_three_derived_levels():
    def count_diff(colors, texts):
        changes = 0
        if (colors[0] != colors[-1]): changes += 1
        if (texts[0] != texts[-1]): changes += 1
        return changes

    def make_k_diff_level(k):
        def k_diff(colors, texts):
            return count_diff(colors, texts) == k
        return DerivedLevel(str(k), Transition(k_diff, [color, text]))

    changed = Factor("changed", [make_k_diff_level(0),
                                 make_k_diff_level(1),
                                 make_k_diff_level(2)]);

    block = CrossBlock([color, text, changed], [color, text], [Exclude(changed["1"])])

    assert block.build_variable_lists((changed, changed["0"])) == [[17, 20, 23]]
    # assert block.build_variable_lists((changed, changed["1"])) == [[18, 21, 24]]
    assert block.build_variable_lists((changed, changed["2"])) == [[19, 22, 25]]

def test_crossing_size_with_complex_excludes():
    deviantColor             = Factor("deviant color",  ["pink", "purple"])
    deviantOrientation       = Factor("deviant orientation", ["left", "right"])
    deviantMovement          = Factor("deviant movement", ["vertical", "horizontal"])
    
    deviantColorObject          = Factor("color deviant", ["object 1", "object 2", "object 3", "object 4"])
    deviantOrientationObject    = Factor("orientation deviant", ["object 1", "object 2", "object 3", "object 4"])
    deviantMovementObject       = Factor("movement deviant", ["object 1", "object 2", "object 3", "object 4"])

    def legalObjectConfiguration(deviantColorObject, deviantOrientationObject, deviantMovementObject):
        return (deviantColorObject != deviantOrientationObject) and (deviantColorObject != deviantMovementObject) and (deviantOrientationObject != deviantMovementObject)
    def illegalObjectConfiguration(deviantColorObject, deviantOrientationObject, deviantMovementObject):
        return not legalObjectConfiguration(deviantColorObject, deviantOrientationObject, deviantMovementObject)
    legalObject = DerivedLevel("legal", WithinTrial(legalObjectConfiguration, [deviantColorObject, deviantOrientationObject, deviantMovementObject]))
    illegalObject = DerivedLevel("illegal", WithinTrial(illegalObjectConfiguration, [deviantColorObject, deviantOrientationObject, deviantMovementObject]))
    objectConfiguration = Factor("object configuration", [
        legalObject,
        illegalObject
    ])

    task              = Factor("task", ["color task", "movement task", "orientation task"])

    def A_B_A(tasks):
        return (tasks[-2] == tasks[0]) and (tasks[-2] != tasks[-1])
    def A_B_C(tasks):
        return (tasks[-2] != tasks[0]) and (tasks[-2] != tasks[-1]) and (tasks[-1] != tasks[0])
    def A_A_B(tasks):
        return (tasks[-2] == tasks[-1]) and (tasks[-1] != tasks[0])
    def A_A_A(tasks):
        return (tasks[-2] == tasks[-1]) and (tasks[-1] == tasks[0])
    def A_B_B(tasks):
        return (tasks[-2] != tasks[-1]) and (tasks[-1] == tasks[0])
    def invalid_transition(tasks):
        return A_A_B(tasks) or A_A_A(tasks) or A_B_B(tasks)

    illegalTransition = DerivedLevel("illegalT", Window(invalid_transition, [task], 3, 1))
    task_transition = Factor("task transition", [
        DerivedLevel("A-B-A", Window(A_B_A, [task], 3, 1)),
        DerivedLevel("A-B-C", Window(A_B_C, [task], 3, 1)),
        illegalTransition
    ])

    constraints = [Exclude((objectConfiguration, illegalObject)),
                   Exclude((task_transition, illegalTransition))
                   ]

    design       = [deviantColorObject, deviantOrientationObject, deviantMovementObject, objectConfiguration, task, task_transition]
    crossing     = [deviantColorObject, deviantOrientationObject, deviantMovementObject, task, task_transition]
    block        = CrossBlock(design, crossing, constraints, require_complete_crossing=False)

    assert block.crossing_size() == 144



"""This module provides functionality for making requests to the backend."""


from typing import List

from sweetpea._internal.logic import And, cnf_to_json
from sweetpea._internal.core import Var
from sweetpea._internal.core.generate.utility import GenerationRequest, AssertionType


class LowLevelRequest:
    """Represents an individual low-level request to the backend."""

    comparisons = ['EQ', 'LT', 'GT']  # will we want more comparisons in the future?

    def __init__(self, comparison: str, k: int, variables: List[int]) -> None:
        self.comparison = comparison
        self.k = k
        self.variables = variables
        self.__validate()

    def __validate(self):
        if self.comparison not in self.comparisons:
            raise ValueError('LowLevelRequest.comparison must be one of ' + str(self.comparisons))

        if not isinstance(self.k, int):
            raise ValueError('LowLevelRequest.k must be an integer')
        # TODO - Non-empty list, list containing non-integers

    def to_dict(self):
        """Converts this request to a dict suitable for conversion to json and
        then the backend.
        """

        return {
            'equalityType': self.comparison,
            'k': self.k,
            'booleanValues': self.variables
        }

    def to_generation_request(self):
        return GenerationRequest(AssertionType[self.comparison], self.k, [Var(v) for v in self.variables])

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)


class BackendRequest:
    """Represents a full request to the backend, including CNFs,
    LowLevelRequests, and unigen arguments.
    """

    def __init__(self, fresh: int, cnfs: List[And] = [], ll_requests: List[LowLevelRequest] = []) -> None:
        self.cnfs = list(cnfs)
        self.ll_requests = list(ll_requests)
        self.fresh = fresh
        self.support = -1
        self.solution_count = -1

    def get_cnfs_as_json(self):
        return cnf_to_json(self.cnfs)

    def get_requests_as_json(self):
        return list(map(lambda r: r.to_dict(), self.ll_requests))

    def get_requests_as_generation_requests(self):
        return list(map(lambda r: r.to_generation_request(), self.ll_requests))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)



import pytest

from sweetpea._internal.logic import And
from sweetpea._internal.sampling_strategy.guided import GuidedGen


def test_committed_to_solution():
    assert GuidedGen._GuidedGen__committed_to_solution([
        And([1, 3, 5]),
        And([2, 4, 6])
    ]) == [1, 3, 5, 2, 4, 6]



from sweetpea import *

color = Factor("color", ['red', 'green'])
size = Factor("size", ['real small', 'great big'])

def test_implied_within_trial():
    match = Factor(name="match", initial_levels=[
        DerivedLevel(name="same", window=WithinTrial(predicate=lambda a, b: a[0] == b[0], factors=[color, size])),
        DerivedLevel(name="diff", window=WithinTrial(predicate=lambda a, b: a[0] != b[0], factors=[color, size]))
    ])

    block      = CrossBlock([color, size, match], crossing=[color, size], constraints=[])
    experiments = synthesize_trials(block=block, samples=4)

    assert len(experiments) == 4
    assert len(experiments[0]["color"]) == 4

    for e in experiments:
        for i in range(0, len(e["color"])):
            if (e["color"][i][0] == e["size"][i][0]):
                assert e["match"][i] == "same"
            else:
                assert e["match"][i] == "diff"

def test_implied_window():
    match = Factor(name="match", initial_levels=[
        DerivedLevel(name="same", window=Window(lambda a: a[0] == a[-1], [color], 2, 1)),
        DerivedLevel(name="diff", window=Window(lambda a: a[0] != a[-1], [color], 2, 1))
    ])

    block      = CrossBlock([color, size, match], crossing=[color, size], constraints=[])
    experiments = synthesize_trials(block=block, samples=4)

    assert len(experiments) == 4
    assert len(experiments[0]["color"]) == 4

    for e in experiments:
        assert e["match"][0] == ""
        for i in range(1, len(e["color"])):
            if (e["color"][i] == e["color"][i-1]):
                assert e["match"][i] == "same"
            else:
                assert e["match"][i] == "diff"



"""This module provides functionality for building graphs representing the
relationships between factors in a given design.
"""


from typing import List

import networkx as nx

from sweetpea._internal.primitive import DerivedFactor, Factor


class DesignGraph():
    """Builds a directed graph representing the relationship between all
    factors in the design. Primary intent is to facilitate preventing invalid
    crossings.
    """

    def __init__(self, design: List[Factor]) -> None:
        self.design = design
        self.graph = self.__build_graph(design)

    def __build_graph(self, design: List[Factor]) -> nx.DiGraph:
        g = nx.DiGraph()

        for f in design:
            # Add the factor as a node.
            g.add_node(f.name)

            # Simple factors (not derived) are the leaves.
            if not isinstance(f, DerivedFactor):
                continue

            # Add directed edges between this factor and all factors that it depends on.
            for l in f.levels:
                for depended_on_factor in l.window.factors:
                    g.add_edge(f.name, depended_on_factor.name)

        return g

    def draw(self):
        import matplotlib as plt
        nx.draw(self.graph, with_labels=True)
        plt.show()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)




from typing import Iterable

def argcheck(who, val, pred_or_type, what):
    pred = (lambda v: isinstance(v, pred_or_type)) if isinstance(pred_or_type, type) else pred_or_type
    if not pred(val):
        raise ValueError(f"{who}: expected {what}, given {val}")

def make_islistof(pred_or_type):
    pred = (lambda v: isinstance(v, pred_or_type)) if isinstance(pred_or_type, type) else pred_or_type
    return (lambda l: isinstance(l, Iterable) and all([pred(v) for v in l]))

def make_istuple(preds_or_types):
    def make_pred(pred_or_type):
        return (lambda v: isinstance(v, pred_or_type)) if isinstance(pred_or_type, type) else pred_or_type
    preds = [make_pred(pred_or_type) for pred_or_type in preds_or_types]
    return (lambda l: isinstance(l, tuple) and len(l) == len(preds_or_types) and all([pred(v) for pred, v in zip(preds, l)]))



"""This module provides the various kinds of blocks that can be used to create
a factorial experimental design.
"""

from abc import abstractmethod
from functools import reduce
from itertools import accumulate, combinations, product, repeat, chain
from typing import List, Union, Tuple, Optional, cast, Any, Dict, Set
from math import ceil
from networkx import has_path
import copy

from sweetpea._internal.block import Block
from sweetpea._internal.backend import BackendRequest
from sweetpea._internal.level import get_all_levels
from sweetpea._internal.primitive import (
    DerivedFactor, DerivedLevel, ElseLevel, Factor, SimpleFactor, SimpleLevel, Level,
    HiddenName
)
from sweetpea._internal.logic import to_cnf_tseitin
from sweetpea._internal.base_constraint import Constraint
from sweetpea._internal.design_graph import DesignGraph
from sweetpea._internal.iter import chunk_list
from sweetpea._internal.weight import combination_weight
from sweetpea._internal.argcheck import argcheck, make_islistof
from sweetpea._internal.sample_conversion import convert_sample_from_names_to_objects
from sweetpea._internal.check_mismatch import combinations_mismatched_weights

class MultiCrossBlockRepeat(Block):
    """An internal :class:`.Block` to handle blocks and repeats.
    """

    def __init__(self,
                 design: List[Factor],
                 crossings: List[List[Factor]],
                 constraints: List[Constraint],
                 require_complete_crossing: bool = True):
        who = "MultiCrossBlockRepeat"
        argcheck(who, design, make_islistof(Factor), "list of Factors for design")
        argcheck(who, crossings, make_islistof(make_islistof(Factor)), "list of list of Factors for crossings")
        argcheck(who, constraints, make_islistof(Constraint), "list of Constraints for constraints")
        self._create(who, design, crossings, constraints, require_complete_crossing, None)

    def _create(self,
                who: str,
                design: List[Factor],
                crossings: List[List[Factor]],
                constraints: List[Constraint],
                require_complete_crossing: bool,
                within_block_count: Optional[int]):
        from sweetpea._internal.constraint import Cross, Consistency
        from sweetpea._internal.derivation_processor import DerivationProcessor
        self.orig_design = design
        self.orig_crossings = crossings
        self.orig_constraints = constraints
        design, crossings, replacements = _desugar_factors_with_weights(design, crossings)
        all_constraints = cast(List[Constraint], [Cross(), Consistency()]) + constraints
        all_constraints = _desugar_constraints(all_constraints, replacements)
        super().__init__(design, crossings, all_constraints, require_complete_crossing, who)
        self.crossing_sizes = [self.crossing_size(c) for c in self.crossings]
        self.preamble_sizes = [self._trials_per_sample_for_one_crossing(c) - self.crossing_size(c)
                               for c in self.crossings]
        self.constraints += DerivationProcessor.generate_derivations(self)
        if (not list(filter(lambda c: c.is_complex_for_combinatoric(), self.constraints))
                and not list(filter(lambda f: f.has_complex_window, design))):
            self.complex_factors_or_constraints = False
        if within_block_count:
            if not all([s == self.preamble_sizes[0] for s in self.preamble_sizes]):
                raise RuntimeError("cannot repeat a block with corssings that have different preamble lengths")
            self.within_block_count = within_block_count
            self.within_block_preamble = self.preamble_sizes[0]
        else:
            self.within_block_count = self.trials_per_sample()
        self.__validate(who)

    def __validate(self, who: str):
        self.__validate_crossing(who)

    def __validate_crossing(self, who: str):
        dg = DesignGraph(self.design).graph
        warnings = []
        template = " '{}' depends on '{}'"
        for crossing in self.crossings:
            combos = combinations(crossing, 2)

            for c in combos:
                if has_path(dg, c[0].name, c[1].name):
                    warnings.append(template.format(c[0].name, c[1].name))
                elif has_path(dg, c[1].name, c[0].name):
                    warnings.append(template.format(c[1].name, c[0].name))

        if warnings:
            self.errors.add("WARNING: Dependencies among factors may make the"
                            " crossing unsatisfiable:"
                            + reduce(lambda accum, s: accum + "\n" + s, warnings, ""))

    def __trials_required_for_crossing(self, f: Factor, crossing_size: int) -> int:
        """Given a factor ``f``, and a crossing size, this function will
        compute the number of trials required to fully cross ``f`` with the
        other factors.

        For example, if ``f`` is a transition, it doesn't apply to trial 1. So
        when the ``crossing_size`` is ``4``, we'd actually need 5 trials to
        fully cross with ``f``.

        This is a helper for :class:`.MultipleCrossBlock.trials_per_sample`.
        """
        trial = 0
        counter = 0
        while counter != crossing_size:
            trial += 1
            if f.applies_to_trial(trial):
                counter += 1
        return trial

    def _trials_per_sample_for_crossing(self):
        """Result includes preamble trials."""
        crossing_size = max(map(lambda c: self.crossing_size(c), self.crossings))
        crossing_trials = list(map(lambda c: list(map(lambda f: self.__trials_required_for_crossing(f, crossing_size),
                                                      c)),
                                   self.crossings))
        required_trials = list(map(lambda l: max([0] + l), crossing_trials))
        return max(required_trials)

    def _trials_per_sample_for_one_crossing(self, c: List[Factor]):
        """Result includes preamble trials."""
        crossing_size = self.crossing_size(c)
        return max([0] + list(map(lambda f: self.__trials_required_for_crossing(f, crossing_size), c)))

    def trials_per_sample(self):
        if self._trials_per_sample:
            return self._trials_per_sample
        self._trials_per_sample = max([self.min_trials, self._trials_per_sample_for_crossing()])
        return self._trials_per_sample

    def variables_per_trial(self):
        # Factors with complex windows are excluded because we don't want variables allocated
        # in every trial when the window spans multiple trials.
        if self._variables_per_trial:
            return self._variables_per_trial
        grid_factors = filter(lambda f: not f.has_complex_window, self.act_design)
        self._variables_per_trial = sum([len(factor.levels) for factor in grid_factors])
        return self._variables_per_trial

    def grid_variables(self):
        return self.trials_per_sample() * self.variables_per_trial()

    def __count_exclusions(self, crossing):
        """This method is responsible for determining the number of trials that
        should be excluded from the full crossing, based on any
        :class:`Exclude` constraints that the user provides, as well as combinations
        that are impossible based on a derived level's definition. A single
        :class:`Exclude` constraint may prevent multiple crossings, depending
        on the derivation function used.
        """
        from sweetpea._internal.constraint import Exclude

        excluded_crossings = cast(Set[Tuple[Level, ...]], set())

        # Generate the full crossing as a list of tuples.
        levels_lists = [list(f.levels) for f in crossing]
        all_crossings = list(product(*levels_lists))

        # Get the exclude constraints.
        exclusions = list(filter(lambda c: isinstance(c, Exclude), self.constraints))

        # Check for impossible combinations
        for c in all_crossings:
            for l in c:
                if isinstance(l, DerivedLevel):
                    f = l.factor
                    if isinstance(f, DerivedFactor) and not f.has_complex_window:
                        argss = []
                        for af in l.window.factors:
                            if af in crossing:
                                # Find level in `c`:
                                for al in c:
                                    if al in af.levels:
                                        argss.append([al.name])
                                        break
                            else:
                                # We'll need to try all possible levels in `af`
                                argss.append([ll.name for ll in af.levels])
                        all_possible_argss = list(product(*argss))
                        if not any([l.window.predicate(*args) for args in all_possible_argss]):
                            excluded_crossings.add(tuple(c))

        # Check for excluded combinations
        for constraint in exclusions:
            # Retrieve the derivation function that defines this exclusion.
            excluded_level = constraint.level

            if excluded_level.factor in crossing:
                for c in all_crossings:
                    if excluded_level in c:
                        excluded_crossings.add(tuple(c))

            if isinstance(excluded_level, SimpleLevel):
                # nothing more to do
                pass
            elif constraint.factor.has_complex_window:
                # We are not obliged to filter impossible cases for a complex level
                continue
            elif excluded_level.factor not in crossing:
                # For each crossing, ensure that at least one combination is possible with the design-only
                # factor, keeping in mind the exclude contraints.
                for c in all_crossings:
                    if all(map(lambda d: self.__excluded_derived(excluded_level, c + d),
                               list(product(*[list(f.levels) for f in filter(lambda f: f not in crossing,
                                                                             self.act_design)])))):
                        excluded_crossings.add(tuple(c))

        if len(excluded_crossings) != 0:
            if self.require_complete_crossing:
                er = "Complete crossing unsatisfiable"
            else:
                er = "WARNING: crossing incomplete"
            er += " due to excluded or impossible combinations:"
            for c in excluded_crossings:
                names = ', '.join([f"'{l.name}'" for l in c])
                er += "\n " + names
            self.errors.add(er)

        return sum([combination_weight(c) for c in excluded_crossings])

    def __excluded_derived(self, excluded_level, c):
        """Given the complete crossing and an exclude constraint, returns true
        if that combination results in the exclude level or if the combination
        is not possible based on the level's definition.
        """
        ret = []

        cx = {l.factor: l.name for l in c}

        for f in filter(lambda f: isinstance(f, DerivedFactor), excluded_level.window.factors):
            if self.__excluded_derived(cx[f], c):
                return True

        # Invoking the predicate this way is only ok because we only do this for WithinTrial windows.
        # With complex windows, it wouldn't work due to the list aspect for each argument.
        return excluded_level.window.predicate(*[cx[f] for f in excluded_level.window.factors])

    def __select_crossing(self, crossing: Optional[List[Factor]]) -> List[Factor]:
        if not crossing:
            if len(self.crossings) != 1:
                raise ValueError("Not a single-crossing block, so crossing must be provided to crossing_size")
            crossing = self.crossings[0]
        return crossing

    def crossing_size(self, crossing: Optional[List[Factor]] = None):
        """The crossing argument must be one of the block's crossings."""
        crossing = self.__select_crossing(crossing)
        crossing_size = self.crossing_size_without_exclusions(crossing)
        crossing_size -= self.__count_exclusions(crossing)
        return crossing_size

    def crossing_size_without_exclusions(self, crossing: List[Factor]):
        """The crossing argument must be one of the block's crossings."""
        return reduce(lambda sum, factor: sum * factor.level_weight_sum(), crossing, 1)

    def preamble_size(self, crossing: Optional[List[Factor]] = None):
        crossing = self.__select_crossing(crossing)
        return self._trials_per_sample_for_one_crossing(crossing) - self.crossing_size(crossing)

    def crossing_weight(self, crossing: Optional[List[Factor]] = None):
        crossing = self.__select_crossing(crossing)
        crossing_size = self.crossing_size(crossing)
        preamble_size = self.preamble_size(crossing)
        return ((self.within_block_count - preamble_size) + (crossing_size - 1)) // crossing_size

    def draw_design_graph(self):
        dg = DesignGraph(self.design)
        dg.draw()

    def factor_in_crossing(self, factor):
        return any(list(map(lambda c: factor in c, self.crossings)))

    def factor_used_in_crossing(self, factor):
        return any(list(map(lambda c: any(list(map(lambda f: f.uses_factor(factor), c))),
                            self.crossings)))

    def sample_mismatch_factors(self, sample: dict) -> list:
        """Test if the factors in a given sequence meet the criteria defined for this factor

        For example in a stroop experiment, if the derived factor congruency is defined as
        equality between the factor word and color, then in the sequence the trials with
        equal word and colors should be labeled congruent.
        """
        res = []
        sample_objects = convert_sample_from_names_to_objects(sample, self.design)
        for factor in self.design:
            if not isinstance(factor.name, HiddenName):
                factor_test = True
                for i in range(len(sample_objects[factor])):
                    factor_test &= factor.test_trial(i, sample_objects)
                if not factor_test:
                    res.append(factor.name)
        return res

    def sample_mismatch_constraints(self, sample: dict) -> list:
        """Test if the factors in a given sequence meet the criteria defined for this constraints"""
        res = []
        sample_objects = convert_sample_from_names_to_objects(sample, self.design)
        for constraint in self.constraints:
            if not constraint.potential_sample_conforms(sample_objects, self):
                pretty_name = constraint.__class__.__name__
                if hasattr(constraint, 'k'):
                    pretty_name += f', {constraint.k}'  # type: ignore
                if hasattr(constraint, 'level'):
                    pretty_name += f', {constraint.level}'  # type: ignore
                res.append(pretty_name)
        return res

    def sample_mismatch_crossing(self, sample: dict, acceptable_error_per_crossing: int = 0) -> list:
        """Test if a given sequence meet the criteria defined for the crossings"""
        sample_objects = convert_sample_from_names_to_objects(sample, self.design)
        res = cast(list, [])
        trial_count = self.trials_per_sample()

        for i, crossing in enumerate(self.crossings):
            bad = 0
            start = self.preamble_sizes[i]
            c_weight = self.crossing_weight(crossing)
            c_crossing_size = self.crossing_sizes[i] * c_weight
            levels_lists = [sample[f.name] for f in crossing]
            # check if length of sample is enough to satisfy the crossings
            for levels in levels_lists:
                if len(levels) != trial_count:
                    res.append(str(crossing))
            while start < trial_count:
                end = start+c_crossing_size
                or_less = False
                if (end > trial_count):
                    end = trial_count
                    or_less = True
                bad += combinations_mismatched_weights(start, end, c_weight, crossing, sample_objects, or_less)
                start += c_crossing_size
            if bad > acceptable_error_per_crossing:
                res.append(str(crossing))
        return res

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return str(self.__dict__)


class MultiCrossBlock(MultiCrossBlockRepeat):
    """A :class:`.Block` with multiple crossings, meant to be used in
    experiment synthesis. Similar to :func:`fully_cross_block`, except it can
    be configured with multiple crossings.

    :param design:
        A :class:`list` of all the :class:`Factors <.Factor>` in the design.
        When a sequence of trials is generated, each trial will have one level
        from each factor in ``design``.

    :param crossings:
        A :class:`list` of :class:`lists <list>` of :class:`Factors <.Factor>`
        representing crossings. The number of trials in each run of the
        experiment is determined by the *maximum* product among the number of
        levels in the crossings.

        Every combination of levels in each individual crossing in
        ``crossings`` appears at least once. Different crossings can refer to
        the same factors, which constrains how factor levels are chosen across
        crossings.

    :param constraints:
        A :class:`list` of :class:`Constraints <.Constraint>` that restrict the
        generated trials.

    :param require_complete_crossing:
        Whether every combination in ``crossing`` must appear in a block of
        trials. ``True`` by default. A ``False`` value is appropriate if
        combinations are excluded through an :class:`.Exclude`
        :class:`.Constraint`.
    """

    def __init__(self,
                 design: List[Factor],
                 crossings: List[List[Factor]],
                 constraints: List[Constraint],
                 require_complete_crossing: bool = True):
        who = "MultiCrossBlock"
        argcheck(who, design, make_islistof(Factor), "list of Factors for design")
        argcheck(who, crossings, make_islistof(make_islistof(Factor)), "list of list of Factors for crossings")
        argcheck(who, constraints, make_islistof(Constraint), "list of Constraints for constraints")
        self._create(who, design, crossings, constraints, require_complete_crossing, None)

class CrossBlock(MultiCrossBlock):
    """A fully crossed :class:`.Block` meant to be used in experiment
    synthesis. This is the preferred mechanism for describing an experiment.

    :param design:
        A :class:`list` of all the :class:`Factors <.Factor>` in the design.
        When a sequence of trials is generated, each trial will have one level
        from each factor in ``design``.

    :param crossing:
        A :class:`list` of :class:`Factors <.Factor>` used to produce
        crossings. The number of trials in each run of the experiment is
        determined as the product of the number of levels of factors in
        ``crossing``.

        If ``require_complete_crossing`` is ``False``, the ``constraints`` can
        reduce the total number of trials.

        Different trial sequences of the experiment will have different
        combinations of levels in different orders. The factors in ``crossing``
        supply an implicit constraint that every combination of levels in the
        cross should appear once. Derived factors impose additional
        constraints: only combinations of levels that are consistent with
        derivations can appear as a trial. Additional constraints can be
        manually imposed via the ``constraints`` parameter.

    :param constraints:
        A :class:`list` of :class:`Constraints <.Constraint>` that restrict the
        generated trials.

    :param require_complete_crossing:
        Whether every combination in ``crossing`` must appear in a block of
        trials. ``True`` by default. A ``False`` value is appropriate if
        combinations are excluded through an :class:`.Exclude`
        :class:`.Constraint`.
    """

    def __init__(self,
                 design: List[Factor],
                 crossing: List[Factor],
                 constraints: List[Constraint],
                 require_complete_crossing: bool = True):
        who = "CrossBlock"
        argcheck(who, design, make_islistof(Factor), "list of Factors for design")
        argcheck(who, crossing, make_islistof(Factor), "list of Factors for crossing")
        argcheck(who, constraints, make_islistof(Constraint), "list of Constraints for constraints")
        self._create(who, design, [crossing], constraints, require_complete_crossing, None)

class Repeat(MultiCrossBlockRepeat):
    def __init__(self,
                 block: MultiCrossBlock,
                 constraints: List[Constraint]):
        from sweetpea._internal.constraint import Exclude
        who = "Repeat"
        
        argcheck(who, block, MultiCrossBlock, "MultiCrossBlock object")
        argcheck(who, constraints, make_islistof(Constraint), "list of Constraints for constraints")

        # We might need another subtype layer in the future, but currently `Exclude`
        # is the only disallowed `Constraint` type:
        for c in constraints:
            if isinstance(c, Exclude):
                raise ValueError("Exclude constraints no allowed in list of constraints")

        block_constraints = [copy.copy(c) for c in block.orig_constraints]
        for c in block_constraints:
            c.set_within_block()

        self._create(who,
                     block.orig_design, block.orig_crossings, block_constraints + constraints,
                     block.require_complete_crossing,
                     block.within_block_count)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~                         Helper functions                            ~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def _desugar_factors_with_weights(design: List[Factor],
                                  crossings: List[List[Factor]]) -> Tuple[List[Factor],
                                                                          List[List[Factor]],
                                                                          dict]:
    # When a derived factor has weighted levels and is in the
    # crossing, then the weight have to be handed by sampling, because
    # it doesn't work to have multiple levels in a derived factor that
    # match the same cases. If a derived factor is not in the
    # crossing, the weights are irrelevnt, because other factors
    # chosen for a combination determine a derived level.
    #
    # For a non-derived factor, weighting is effecively the same as
    # having multiple levels with the same name. Still, as long as a
    # factor with weights is used in the crossing (all of them, in the
    # case of multiple crossings), then we leave the weights in place
    # and handle them in sampling.
    #
    # But when a non-derived factor with weights is not in (all of
    # the) crossing(s), we desugar to a factor with multiple levels
    # that have the same name. That makes the biasing effect of
    # weighting work for formula-based samplers, and it geneally means
    # that samplers do not have to handle the weights specifically.
    #
    # To desugar, we create new factors and levels, and we rewrite all
    # constraints and derived factors to refer to the new ones. Each
    # desugared factor has two replacements: a non-derived factors
    # with the weights turned into multiple levels, and a derived
    # factor that has the same level names as before. The derived
    # factor is needed in case a constraint refers to an weighted
    # level that gets expanded in the non-derived factor.
    #
    # The `replacements` dictionary maps a level to its replacement,
    # and it maps factor to a list of two factors: the derived
    # replacement and non-derived replacement.
    #
    weighted = []
    for f in design:
        if (not isinstance(f, DerivedFactor)) and any([l.weight > 1 for l in f.levels]):
            if all([not f in c for c in crossings]):
                weighted.append(f)
    if not weighted:
        # No desugaring needed
        return (design, crossings, {})
    else:
        # Desugaring needed
        replacements = cast(dict, {})
        for f in weighted:
            # Adds to `replacements`:
            cast(SimpleFactor, f).desugar_weights(replacements)
        for f in design:
            if isinstance(f, DerivedFactor):
                # Uses `replacements`:
                f.desugar_for_weights(replacements)
        # Returned `replacements` is also used for constraint desugaring
        return (list(chain.from_iterable([replacements.get(f, [f]) for f in design])),
                [[replacements.get(f, [f, f])[1] for f in c] for c in crossings],
                replacements)


def _desugar_constraints(constraints: List[Constraint], replacements: dict) -> List[Constraint]:
    desugared_constraints = []
    for c in constraints:
        desugared_constraints.extend(c.desugar(replacements))
    return desugared_constraints



"""This module makes it easy to set up the Unigen and CryptoMiniSAT executables
as needed by Sweetpea. It can also be run as a script.

.. note::

    SweetPea Core only makes use of the automated download of the current
    system and machine-type's executables from the latest release. The rest of
    the code accommodates other use cases and is not strictly necessary, but it
    took a bit of reading to suss out GitHub's API so I figured I'd leave it in
    place for future use if necessary. But perhaps it should be removed at some
    point.

Using This Module As a Script
=============================

To use this module as a script, do ``python executables.py`` (specifying the
full path as needed). A number of options are supported:

-h, --help
    Show a help message and exit.

-d BIN_DIR, --bin-dir BIN_DIR
    The directory into which to install the executables.

    The default is determined by :func:`appdirs.user_data_dir` +
    ``SweetPea/Executables``.

-s SYSTEM, --system SYSTEM
    The target system. Valid options are ``None``, ``Darwin``, ``Linux``, and
    ``Windows``. When ``None`` is given, the system will be automatically
    deduced by :func:`platform.system`.

    The default is ``None``.

-m MACHINE, --machine MACHINE
    The target machine type. Valid options are ``None``, ``arm64``, ``x86_64``,
    and ``AMD64``. When ``None`` is given, the machine type will be
    automatically deduced by :func:`platform.machine`.

    The default is ``None``.

-t TAG, --tag TAG
    The `sweetpea-org/unigen-exe <https://github.com/sweetpea-org/unigen-exe>`_
    release tag to target for downloading.

    The default is to use the latest available release tag.

--asset-string
    Prints out the asset string for the indicated system+machine combination.
    This is provided mostly for debugging.
"""


import platform
import stat

from appdirs import user_data_dir
from itertools import groupby
from json import loads as load_json
from os import chmod, environ
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Iterator, List, Optional, Tuple
from urllib.request import Request, urlopen
from zipfile import ZipFile, ZipInfo


__all__ = [
    # We export the environment variable names so they will be documented.
    'DOWNLOAD_UNIGEN_ENV_VAR', 'UNIGEN_EXE_ENV_VAR',
    # The rest of the exports are for use in other modules.
    'CRYPTOMINISAT_EXE', 'DEFAULT_DOWNLOAD_IF_MISSING', 'UNIGEN_EXE', 'CMSGEN_EXE',
    'ensure_executable_available'
]


JSONDict = Dict[str, Any]

#: The name of the environment variable that can be used to specify whether the
#: bundled executables should be downloaded. The default value is ``True``.
#: Valid options are:
#:
#: ===============================  ================================================================
#: Download executables if missing  ``True``, ``true``, ``T``, ``t``, ``Yes``, ``yes``, ``Y``, ``y``
#: Do not download executables      ``False``, ``false``, ``F``, ``f``, ``No``, ``no``, ``N``, ``n``
#: ===============================  ================================================================
DOWNLOAD_UNIGEN_ENV_VAR = 'UNIGEN_DOWNLOAD_IF_MISSING'
if DOWNLOAD_UNIGEN_ENV_VAR in environ:
    download_if_missing = environ[DOWNLOAD_UNIGEN_ENV_VAR]
    if download_if_missing in ('True', 'true', 'T', 't', 'Yes', 'yes', 'Y', 'y'):
        DEFAULT_DOWNLOAD_IF_MISSING = True
    elif download_if_missing in ('False', 'false', 'F', 'f', 'No', 'no', 'N', 'n'):
        DEFAULT_DOWNLOAD_IF_MISSING = False
    else:
        raise RuntimeError(f"Invalid {DOWNLOAD_UNIGEN_ENV_VAR} value: {download_if_missing}. Please use 'True' or 'False'.")
else:
    DEFAULT_DOWNLOAD_IF_MISSING = True

UNIGEN_EXE_LATEST_RELEASE_URL = "https://api.github.com/repos/sweetpea-org/unigen-exe/releases/latest"
UNIGEN_EXE_SPECIFIC_TAG_URL = "https://api.github.com/repos/sweetpea-org/unigen-exe/releases/tags/{tag}"

_ASSET_NAMES = {
    ('Darwin',  'arm64'):  'mac-apple-silicon',
    ('Darwin',  'x86_64'): 'mac-intel',
    ('Linux',   'x86_64'): 'linux-x86_64',
    ('Windows', 'x86_64'): 'win-x64',  # TODO: Verify that this is needed.
    ('Windows', 'AMD64'):  'win-x64',
}

_VALID_ASSETS = {k: list(p[1] for p in v) for k, v in groupby(_ASSET_NAMES.keys(), lambda p: p[0])}

#: The folder in which executables will be stored if they are going to be
#: downloaded. The default is an automatically generated directory in the
#: user's data directory as determined by :func:`appdirs.user_data_dir`.
UNIGEN_EXE_ENV_VAR = 'UNIGEN_EXE_DIR'
if UNIGEN_EXE_ENV_VAR in environ:
    EXE_BIN_LOCATION = Path(environ[UNIGEN_EXE_ENV_VAR])
else:
    EXE_BIN_LOCATION = Path(user_data_dir('SweetPea', 'SweetPea-Org')) / 'Executables'


def _build_exe_name(base_name: str) -> Path:
    path = EXE_BIN_LOCATION / base_name
    if platform.system() == 'Windows':
        path = path.with_suffix('.exe')
    return path


# The various executables we provide.
APPROXMC_EXE = _build_exe_name('approxmc')
CRYPTOMINISAT_EXE = _build_exe_name('cryptominisat5')
UNIGEN_EXE = _build_exe_name('unigen')
CMSGEN_EXE = _build_exe_name('cmsgen')

# An easy-to-use collection of the paths.
EXECUTABLE_PATHS = [
    APPROXMC_EXE,
    CRYPTOMINISAT_EXE,
    UNIGEN_EXE,
    CMSGEN_EXE,
]


def _select_asset_for_host_platform() -> Tuple[str, str]:
    """Select the correct asset for the host platform."""
    system = platform.system()
    machine = platform.machine()
    if system not in _VALID_ASSETS:
        raise RuntimeError(f"Unsupported system: {system}")
    if machine not in _VALID_ASSETS[system]:
        raise RuntimeError(f"Unsupported {system} machine type: {machine}")
    return system, machine


def _get_asset_path(system: Optional[str], machine: Optional[str]) -> str:
    """Retrieves the name of the appropriate release asset."""
    if system is None and machine is None:
        return _ASSET_NAMES[_select_asset_for_host_platform()]
    elif system is not None and machine is not None:
        return _ASSET_NAMES[(system, machine)]
    else:
        # We don't allow one or the other --- it must be both or neither.
        raise RuntimeError(f"Must specify (a) both system and machine or (b) neither. Got system: {system}, machine: {machine}")


def _get_asset_zip(system: Optional[str], machine: Optional[str]) -> str:
    """Retrieves the ``.zip`` name of the appropriate release asset."""
    return _get_asset_path(system, machine) + '.zip'


def _get_release_from_url(url: str) -> List[JSONDict]:
    """Queries the GitHub API to get the assets corresponding to a release
    specified by the given URL.
    """
    response = urlopen(url)
    response_bytes = response.read()
    response_text = response_bytes.decode()
    response_json = load_json(response_text)
    return [asset for asset in response_json['assets']]


def get_latest_release_assets() -> List[JSONDict]:
    """Returns a list of JSON dictionaries, each corresponding to an asset in
    the latest release of the `sweetpea-org/unigen-exe repository
    <https://github.com/sweetpea-org/unigen-exe>`_.
    """
    return _get_release_from_url(UNIGEN_EXE_LATEST_RELEASE_URL)


def get_specific_release(tag: str) -> List[JSONDict]:
    """Returns a list of JSON dictionaries, each corresponding to an asset in
    the indicated release of the `sweetpea-org/unigen-exe repository
    <https://github.com/sweetpea-org/unigen-exe>`_..
    """
    url = UNIGEN_EXE_SPECIFIC_TAG_URL.format(tag=tag)
    return _get_release_from_url(url)


def ensure_dir_path_exists(path: Path):
    """Make sure the specified path is a directory and that it and all its
    parents exist.
    """
    path.mkdir(parents=True, exist_ok=True)


def get_asset_url_for_release(system: Optional[str] = None,
                              machine: Optional[str] = None,
                              tag: Optional[str] = None
                              ) -> str:
    """Produces the URL corresponding to the indicated asset for the
    appropriate release. If ``system`` and ``machine`` are unspecified, they
    will be deduced automatically from the host platform's self-report
    information. If the ``tag`` is unspecified, the latest release will be
    used.
    """
    # First resolve the needed asset name.
    asset_zip_name = _get_asset_zip(system, machine)
    # Then we obtain the appropriate list of assets for the correct release.
    if tag is None:
        release_assets = get_latest_release_assets()
    else:
        release_assets = get_specific_release(tag)
    # Now we search the assets for our desired asset.
    matches = [asset for asset in release_assets if asset['name'] == asset_zip_name]
    if len(matches) != 1:
        raise RuntimeError(f"Did not find exactly one asset named {asset_zip_name} for "
                           f"{'release on tag ' + tag if tag is not None else 'latest release'}.")
    asset = matches[0]
    asset_url = asset['url']
    return asset_url


def download_and_extract_asset_zip_for_release(to_bin_dir: Path,
                                               system: Optional[str] = None,
                                               machine: Optional[str] = None,
                                               tag: Optional[str] = None
                                               ) -> Iterator[Path]:
    """Fetches the ``.zip`` archive corresponding to the indicated system and
    machine from the indicated tag, or else use the automatically deduced
    system and machine from the latest release.
    """
    asset_url = get_asset_url_for_release(system, machine, tag)
    ensure_dir_path_exists(to_bin_dir)
    request = Request(asset_url)
    request.add_header('Accept', 'application/octet-stream')
    # We first have to download the .zip archive to a temporary file.
    # NOTE: I did try to implement this without the intermediate save-to-a-
    #       temp-file step, but the ZipFile initializer seems to not like
    #       accepting the url_file as an argument (even though I think the
    #       latter is a "file-like object" and ZipFile claims it accepts
    #       those).
    with NamedTemporaryFile(suffix='.zip') as temp_file:
        temp_file.close()
        temp_file_path = Path(temp_file.name)
        with urlopen(request) as url_file:
            temp_file_path.write_bytes(url_file.read())
        # With the file downloaded, we can extract its contents.
        zf = ZipFile(temp_file_path)
        internal_base_asset_path = _get_asset_path(system, machine) + '/'
        write_buffer: List[Tuple[ZipInfo, Path]] = []
        for zipinfo in zf.infolist():
            if zipinfo.filename == internal_base_asset_path:
                # We don't do anything with the base path.
                continue
            if zipinfo.filename.startswith(internal_base_asset_path):
                # We do a reprocessing pass to ensure none of the paths we
                # want to write to are already occupied. If they are, we refuse
                # to extract anything.
                sub_path = Path(zipinfo.filename).relative_to(internal_base_asset_path)
                new_path = to_bin_dir / sub_path
                if new_path.is_file():
                    # We'll need to do something different here one day when we
                    # want to update existng executables...
                    print(f"File exists; no contents extracted from zip: {new_path}")
                else:
                    write_buffer.append((zipinfo, new_path))
            else:
                # We only extract the needed contents.
                continue
        # Now we go through and actually perform the extractions.
        for zipinfo, new_path in write_buffer:
            new_path.write_bytes(zf.read(zipinfo))
            yield new_path


def ensure_executable_permissions(executable_path: Path):
    """Set permissions on the indicated executable such that the file owner can
    read, write, or execute it, and the owner's group and other users can read
    and execute it.
    """
    # chmod 755 executable_path
    chmod(executable_path,
          stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)


def download_executables(to_bin_dir: Optional[Path] = None,
                         system: Optional[str] = None,
                         machine: Optional[str] = None,
                         tag: Optional[str] = None):
    """Performs the download and extraction of the bundled executables."""
    if to_bin_dir is None:
        to_bin_dir = EXE_BIN_LOCATION
    print(f"Downloading and extracting SweetPea executables to {to_bin_dir}...")
    for extracted_path in download_and_extract_asset_zip_for_release(to_bin_dir, system, machine, tag):
        print(f"    {extracted_path}")
    for executable_path in EXECUTABLE_PATHS:
        ensure_executable_permissions(executable_path)
    print("Done.")


def ensure_executable_available(executable_path: Path, download_if_missing: bool = DEFAULT_DOWNLOAD_IF_MISSING):
    """Checks whether a given necessary executable is available and, if not,
    downloads it (and its companions) automatically.
    """
    if executable_path in EXECUTABLE_PATHS and not executable_path.is_file():
        if download_if_missing:
            download_executables()
        else:
            raise RuntimeError(f"Could not find binary: {executable_path}")


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--bin-dir', type=Path, default=EXE_BIN_LOCATION,
                        help=f"the directory to install executables to; default is: {EXE_BIN_LOCATION}")
    parser.add_argument('-s', '--system', choices=['None'] + [p[0] for p in _ASSET_NAMES.keys()], default='None',
                        help="the target system; default is determined platform.system()")
    parser.add_argument('-m', '--machine', choices=['None'] + [p[1] for p in _ASSET_NAMES.keys()], default='None',
                        help="the target machine type; default is determined by platform.machine()")
    parser.add_argument('-t', '--tag', default=None,
                        help="the unigen-exe tag to target; default is the latest tag available")
    parser.add_argument('--asset-string', action='store_true',
                        help="only print out the asset string for the indicated system+machine combination")
    args = parser.parse_args()

    if args.system == 'None':
        args.system = None
    if args.machine == 'None':
        args.machine = None

    if args.asset_string:
        print(_get_asset_path(system=args.system, machine=args.machine))
    else:
        download_executables(
            to_bin_dir=args.bin_dir,
            system=args.system,
            machine=args.machine,
            tag=args.tag)



from sweetpea import *

def test_window_with_early_start():
    color = Factor("color", ["red", "green", "blue"])
    flavor = Factor("flavor", ["mild", "spicy"])
    redness = Factor("redness",
                     [DerivedLevel("ruddy", Window(lambda c: (c[-1] != "red" and c[0] == "red"), [color], 2, 1, 0)),
                      ElseLevel("plain")])

    block = CrossBlock([color, flavor, redness], [flavor, redness], [])

    i_experiments = synthesize_trials(block, 100, IterateSATGen)
    r_experiments = synthesize_trials(block, 100, RandomGen)

    assert len(i_experiments) == 64
    assert len(r_experiments) == 64

    distinct = set()
    for i in i_experiments:
        distinct.add(tuple([(k, tuple(i[k])) for k in ["color", "flavor", "redness"]]))
    for i in r_experiments:
        assert tuple([(k, tuple(i[k])) for k in ["color", "flavor", "redness"]]) in distinct
    assert len(distinct) == 64

    color = Factor("color", ["red"])

def test_window_with_late_start():
    color = Factor("color", ["red"])
    redness = Factor("redness",
                     [DerivedLevel("ruddy", Window(lambda c: c == "red", [color], 1, 1, 10))])

    block = CrossBlock([color, redness], [color, redness], [])

    i_experiments = synthesize_trials(block, 20, IterateSATGen)
    r_experiments = synthesize_trials(block, 20, RandomGen)
    assert len(i_experiments) == 1
    assert len(r_experiments) == 1

def test_window_early_start_regression():
    letter = Factor('letter', ['b', 'f', 'm', 'q', 'k', 'x', 'r', 'h'])

    def is_target(letters):
        return letters[-1] == letters[0]

    def is_not_target(letters):
        return not is_target(letters)

    one_t = DerivedLevel(1, Window(is_target, [letter], width=2, stride=1, start=0), 1)
    two_t = DerivedLevel(2, Window(is_not_target, [letter], width=2, stride=1, start=0), 1)

    target = Factor('target', [one_t, two_t])

    block = Repeat(CrossBlock(design=[letter, target],
                              constraints=[],
                              crossing=[letter, target]),
                   constraints=[MinimumTrials(48)])

    experiments = synthesize_trials(block, 1)

    # Only possibile solutions alternate bewteen 1 and 2
    for i, v in enumerate(experiments[0]['target']):
        assert v == ((i+1) % 2) + 1

def test_window_early_start_regression2():
    letter = Factor('letter', ['b', 'f'])

    def is_target(letters):
        return letters[-2] == letters[0]

    def is_not_target(letters):
        return not is_target(letters)

    one_t = DerivedLevel('1', Window(is_target, [letter], width=3, stride=1, start=0), 1)
    two_t = DerivedLevel('2', Window(is_not_target, [letter], width=3, stride=1, start=0), 2)

    target = Factor('target', [one_t, two_t])
    
    block = CrossBlock(design=[letter, target],
                       constraints=[],
                       crossing=[letter, target])

    i_experiments = synthesize_trials(block, 10, IterateSATGen)
    r_experiments = synthesize_trials(block, 10, RandomGen)
    assert len(i_experiments) == 8
    assert len(r_experiments) == 8



"""This module provides functionality for partitioning designs into subsets of
factors.
"""


from sweetpea._internal.block import Block
from sweetpea._internal.primitive import DerivedFactor


class DesignPartitions():
    """Encapsulates the logic for partitioning a design into specific subsets
    of factors. For example, factors which are in the design, but not in the
    crossing, or not constrained by any derived levels, etc.
    """

    def __init__(self, block: Block) -> None:
        self._block = block
        self._crossed = None

    def get_crossed_noncomplex_factors(self):
        """Only gets factors for the first crossing."""
        if self._crossed:
            return self._crossed
        result = self._block.crossings[0]
        result = list(filter(lambda f: not f.has_complex_window, result))
        self._crossed = result
        return result

    def get_crossed_noncomplex_derived_factors(self):
        return list(filter(lambda f: isinstance(f, DerivedFactor), self.get_crossed_noncomplex_factors()))

    def get_crossed_complex_factors(self):
        """Only gets factors for the first crossing."""
        result = []
        for f in self._block.crossings[0]:
            if f.has_complex_window:
                if f not in result:
                    result.append(f)
        return result

    def get_uncrossed_and_complex_factors(self):
        crossed = self.get_crossed_noncomplex_factors()
        return list(filter(lambda f: f not in crossed, self._block.act_design))

    def get_source_factors(self):
        # Source factors are depended on by at least one noncomplex derived factor in the crossed factors.
        source_factors = []
        for derived_factor in self.get_crossed_noncomplex_derived_factors():
            for source_factor in derived_factor.levels[0].window.factors:
                if source_factor not in source_factors:
                    source_factors.append(source_factor)
        return source_factors

    def get_uncrossed_basic_factors(self):
        uncrossed = self.get_uncrossed_and_complex_factors()
        return list(filter(lambda f: not isinstance(f, DerivedFactor), uncrossed))

    def get_uncrossed_basic_source_factors(self):
        source_factors = self.get_source_factors()
        return list(filter(lambda f: f in source_factors, self.get_uncrossed_basic_factors()))

    def get_uncrossed_basic_independent_factors(self):
        source_factors = self.get_source_factors()
        return list(filter(lambda f: f not in source_factors, self.get_uncrossed_basic_factors()))

    def get_uncrossed_derived_and_complex_derived_factors(self):
        return list(filter(lambda f: isinstance(f, DerivedFactor), self.get_uncrossed_and_complex_factors()))

    def get_basic_factors(self):
        return list(filter(lambda f: not isinstance(f, DerivedFactor), self._block.act_design))

    def get_derived_factors(self):
        return list(filter(lambda f: isinstance(f, DerivedFactor), self._block.act_design))



import operator as op
import pytest

from sweetpea import Factor, DerivedLevel, ElseLevel, WithinTrial, Transition, Window, Level, CrossBlock
from sweetpea import synthesize_trials, RandomGen

color = Factor("color", ["red", "blue"])
text = Factor("text", ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_level   = DerivedLevel("yes", Transition(op.eq, [color]))
color_no_repeat_level = DerivedLevel("no", Transition(op.ne, [color]))
color_repeats_factor  = Factor("color repeats?", [color_repeats_level, color_no_repeat_level])

color3 = Factor("color3", ["red", "blue", "green"])

yes_fn = lambda colors: colors[0] == colors[1] == colors[2]
no_fn = lambda colors: not yes_fn(colors)
color3_repeats_factor = Factor("color3 repeats?", [
    DerivedLevel("yes", Window(yes_fn, [color3], 3, 1)),
    DerivedLevel("no",  Window(no_fn, [color3], 3, 1))
])


def test_factor_validation():
    Factor("factor name", ["level 1", "level 2"])
    Factor("name", [1, 2])

    # Non-string name
    with pytest.raises(ValueError):
        Factor(56, ["level "])

    # Non-list levels
    with pytest.raises(TypeError):
        Factor("name", 42)

    # Empty list
    with pytest.raises(ValueError):
        Factor("name", [])

    # Duplicate level names
    with pytest.raises(ValueError):
        Factor("name", ["level1", "level1"])
    with pytest.raises(ValueError):
        Factor("name", [Level("level1"), Level("level1")])

    # Valid level types, but not uniform.
    with pytest.raises(ValueError):
        Factor("name", ["level1", con_level])

    # Derived levels with non-uniform window types
    with pytest.raises(ValueError):
        Factor("name", [
            con_level,
            DerivedLevel("other", WithinTrial(lambda color: color, [color]))
        ])

    # Derived levels with different window arguments
    with pytest.raises(ValueError):
        Factor("name", [
            con_level,
            DerivedLevel("other", Transition(lambda colors: colors[0] == colors[1], [color]))
        ])


def test_factor_get_level():
    assert color['red'].name == 'red'
    assert color_repeats_factor['yes'] == color_repeats_level
    assert color.get_level("bogus") is None


def test_factor_has_complex_window():
	assert color.has_complex_window == False
	assert con_factor.has_complex_window == False
	assert color_repeats_factor.has_complex_window == True


def test_factor_applies_to_trial():
    assert color.applies_to_trial(1) == True
    assert color.applies_to_trial(2) == True
    assert color.applies_to_trial(3) == True
    assert color.applies_to_trial(4) == True

    with pytest.raises(ValueError):
        color.applies_to_trial(0)

    assert color_repeats_factor.applies_to_trial(1) == False
    assert color_repeats_factor.applies_to_trial(2) == True
    assert color_repeats_factor.applies_to_trial(3) == True
    assert color_repeats_factor.applies_to_trial(4) == True

    f = Factor('f', [DerivedLevel('l', Window(op.eq, [color], 2, 2))])
    assert f.applies_to_trial(1) == False
    assert f.applies_to_trial(2) == True
    assert f.applies_to_trial(3) == False
    assert f.applies_to_trial(4) == True

    assert color3.applies_to_trial(1) == True


def test_derived_level_validation():
    DerivedLevel(42, WithinTrial(op.eq, [color, text]))
    DerivedLevel("name", WithinTrial(lambda x: x, [color]))

    # Invalid Window
    with pytest.raises(TypeError):
        DerivedLevel("name", 42)

def test_derived_level_argument_list_expansion():
    # Used to internally duplicate each factor to match the width of the window, but not any more
    assert color_repeats_level.window.factors == [color]
    assert color_no_repeat_level.window.factors == [color]


def test_derived_level_get_dependent_cross_product():
    assert [((tup[0].factor.name, tup[0].name),
             (tup[1].factor.name, tup[1].name))
            for tup in con_level.get_dependent_cross_product()] == [
        (('color', 'red'), ('text', 'red')),
        (('color', 'red'), ('text', 'blue')),
        (('color', 'blue'), ('text', 'red')),
        (('color', 'blue'), ('text', 'blue'))]

    integer = Factor("integer", ["1", "2"])
    numeral = Factor("numeral", ["I", "II"])
    text = Factor("text", ["one", "two"])
    two_con_level = DerivedLevel("twoCon", WithinTrial(lambda x: x, [integer, numeral, text]))
    assert [((tup[0].factor.name, tup[0].name),
             (tup[1].factor.name, tup[1].name),
             (tup[2].factor.name, tup[2].name))
            for tup in two_con_level.get_dependent_cross_product()] == [
        (('integer', '1'), ('numeral', 'I'), ('text', 'one')),
        (('integer', '1'), ('numeral', 'I'), ('text', 'two')),
        (('integer', '1'), ('numeral', 'II'), ('text', 'one')),
        (('integer', '1'), ('numeral', 'II'), ('text', 'two')),
        (('integer', '2'), ('numeral', 'I'), ('text', 'one')),
        (('integer', '2'), ('numeral', 'I'), ('text', 'two')),
        (('integer', '2'), ('numeral', 'II'), ('text', 'one')),
        (('integer', '2'), ('numeral', 'II'), ('text', 'two'))]

    mixed_level = DerivedLevel("mixed", WithinTrial(lambda x: x, [
        Factor("color", ["red", "blue", "green"]),
        Factor("boolean", ["true", "false"])
    ]))
    assert [((tup[0].factor.name, tup[0].name),
             (tup[1].factor.name, tup[1].name))
            for tup in mixed_level.get_dependent_cross_product()] == [
        (('color', 'red'), ('boolean', 'true')),
        (('color', 'red'), ('boolean', 'false')),
        (('color', 'blue'), ('boolean', 'true')),
        (('color', 'blue'), ('boolean', 'false')),
        (('color', 'green'), ('boolean', 'true')),
        (('color', 'green'), ('boolean', 'false'))
    ]


def test_derived_level_equality():
    assert con_level == con_level

def test_else_level():
    color = Factor("color", ["red", "blue"])
    text  = Factor("text",  ["red", "blue"])

    conLevel  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
    conFactor = Factor("congruent?", [conLevel, ElseLevel("inc")])

def __get_response_transition() -> Factor:
    color  = Factor("color",  ["red", "blue", "green"])
    motion = Factor("motion", ["up", "down"])
    task   = Factor("task",   ["color", "motion"])

    # Response Definition
    def response_left(task, color, motion):
        return (task == "color"  and color  == "red") or \
            (task == "motion" and motion == "up")

    def response_right(task, color, motion):
        return not response_left(task, color, motion)

    response = Factor("response", [
        DerivedLevel("left",  WithinTrial(response_left,  [task, color, motion])),
        DerivedLevel("right", WithinTrial(response_right, [task, color, motion]))
    ])

    return Factor("response transition", [
        DerivedLevel("repeat", Transition(lambda responses: responses[0] == responses[1], [response])),
        DerivedLevel("switch", Transition(lambda responses: responses[0] != responses[1], [response]))
    ])


def test_derived_level_get_dependent_cross_product_with_nesting():
    response_transition = __get_response_transition()

    assert [((tup[0].factor.name, tup[0].name),
             (tup[1].factor.name, tup[1].name))
            for tup in response_transition.levels[0].get_dependent_cross_product()] == [
        (('response', 'left' ), ('response', 'left' )),
        (('response', 'left' ), ('response', 'right')),
        (('response', 'right'), ('response', 'left' )),
        (('response', 'right'), ('response', 'right'))
    ]

def test_base_window_validation():
    # Nonfactor argument
    with pytest.raises(TypeError):
        WithinTrial(op.eq, [42])

    # Duplicated factors
    with pytest.raises(ValueError):
        DerivedLevel("name", WithinTrial(lambda x, y: x, [color, color]))

def test_non_string_level_names():
    number = Factor('number', [1, 2])

    assert number[1].name == 1
    assert number[1] == number.get_level(1)

    assert number[2].name == 2
    assert number[2] == number.get_level(2)

    numbers = Factor('numbers', [(1, 2), (3, 4), (5, 6, 7)])

    assert numbers[(1, 2)].name == (1, 2)
    assert numbers[(1, 2)] == numbers.get_level((1, 2))

    assert numbers[(3, 4)].name == (3, 4)
    assert numbers[(5, 6, 7)].name == (5, 6, 7)

    # Make sure non-string names are supplied to derived-level predicates
    
    def is_one(n): return n == 1
    def is_two(n): return n == 2

    one = DerivedLevel('one', WithinTrial(is_one, [number]))
    two = DerivedLevel('two', WithinTrial(is_two, [number]))

    derived = Factor("derived", [one,two])
    t = synthesize_trials(CrossBlock([number, derived], [derived], []), 1, RandomGen)
    assert len(t) == 1
    



from abc import ABC, abstractmethod
from typing import List, cast
from itertools import repeat

from sweetpea._internal.block import Block
from sweetpea._internal.iter import intersperse


"""
Data object for sampling result.
"""
class SamplingResult:
    def __init__(self, samples: List[dict], metrics: dict) -> None:
        self.samples = samples
        self.metrics = metrics

"""
Generic interface for sampling strategies.
"""
class Gen(ABC):

    """
    Sample some number of trial sequences for the given block.

    TODO: This should accept some kind of options structure. What if we want
    to disable metrics? Or use some other feature flag?
    """
    @staticmethod
    @abstractmethod
    def sample(block: Block, sample_count: int) -> SamplingResult:
        pass

    """
    Decodes a single solution into a dict of this form:

        {
        '<factor name>': ['<trial 1 label>', '<trial 2 label>, ...]
        ...
        }

    For factors that don't have a value for a given level, such as Transitions,
    the label will be ''.
    """
    @staticmethod
    def decode(block: Block, solution: List[int]) -> dict:
        # Sort the list and remove any negative (false) variables
        solution.sort()
        solution = list(filter(lambda v: v > 0, solution))

        # Separate into simple/complex variables.
        simple_variables = list(filter(lambda v: v <= block.grid_variables(), solution))
        complex_variables = list(filter(lambda v: v > block.grid_variables(), solution))

        experiment = cast(dict, {})

        # Simple factors
        tuples = list(map(lambda v: block.decode_variable(v), simple_variables))
        string_tuples = list(map(lambda t: (t[0].name, t[1].name), tuples))
        for (factor_name, level_name) in string_tuples:
            if factor_name not in experiment:
                experiment[factor_name] = []
            experiment[factor_name].append(level_name)

        # Complex factors - The challenge here is knowing when to insert '', rather than using the variables.
        # Start after 'width' trials, and shift 'stride' trials for each variable.
        complex_factors = list(filter(lambda f: f.has_complex_window, block.act_design))
        for f in complex_factors:
            # Get variables for this factor
            start = block.first_variable_for_level(f, f.levels[0]) + 1
            end = start + block.variables_for_factor(f)
            variables = list(filter(lambda n: n in range(start, end), complex_variables))

            # Get the level names for the variables in the solution.
            level_tuples = list(map(lambda v: block.decode_variable(v), variables))
            level_names = list(map(lambda t: (t[1].name), level_tuples))

            # Intersperse empty strings for the trials to which this factor does not apply.
            #level_names = list(intersperse('', level_names, f.levels[0].window.stride - 1))
            #level_names = list(repeat('', f.levels[0].window.width - 1)) + level_names
            level_names_fill = []
            for n in range(block.trials_per_sample()):
                level_names_fill.append(level_names.pop(0) if f.applies_to_trial(n+1) else '')
            experiment[f.name] = level_names_fill

        return experiment

    @staticmethod
    def class_name():
        return "Unknown Sampling Strategy"



"""This module provides functionality to communicate with the server."""


from typing import List

from sweetpea._internal.block import Block
from sweetpea._internal.core import CNF, combine_cnf_with_requests, cnf_is_satisfiable
from sweetpea._internal.logic import And, cnf_to_json


def build_cnf(block: Block) -> CNF:
    """Converts a Block into a CNF represented as a Unigen-compatible string.
    """
    backend_request = block.build_backend_request()
    cnf = CNF(backend_request.get_cnfs_as_json())
    combined_cnf = combine_cnf_with_requests(
        cnf,
        backend_request.fresh - 1,
        block.variables_per_sample(),
        backend_request.get_requests_as_generation_requests())
    return combined_cnf


def is_cnf_still_sat(block: Block, additional_clauses: List[And]) -> bool:
    backend_request = block.build_backend_request()
    cnf = CNF(backend_request.get_cnfs_as_json()) + CNF(cnf_to_json(additional_clauses))
    combined_cnf = combine_cnf_with_requests(
        cnf,
        backend_request.fresh - 1,
        block.variables_per_sample(),
        backend_request.get_requests_as_generation_requests())
    return cnf_is_satisfiable(combined_cnf)



"""This module provides tests for SweetPea's Core framework."""


from operator import invert
from typing import Callable, List, Optional, TypeVar

from .cnf import CNF, Var


T = TypeVar('T')


def sequence(xss: List[List[T]]) -> List[List[T]]:
    if not xss:
        return [[]]
    return [[x] + xs for x in xss[0] for xs in sequence(xss[1:])]


def test_half_adder_dimacs() -> List[str]:
    cnf = CNF()
    a = cnf.get_fresh()
    b = cnf.get_fresh()
    _, _ = cnf.half_adder(a, b)
    all_inputs = sequence([[a, ~a], [b, ~b]])
    test_constraints = [cnf & p[0] & p[1] for p in all_inputs]
    # test_constraints = [cnf + CNF(x[0]) + CNF(x[1]) for x in all_inputs]
    return [cnf.as_unigen_string() for cnf in test_constraints]


def test_full_adder_dimacs() -> List[str]:
    cnf = CNF()
    a, b, c = cnf.get_n_fresh(3)
    _, _ = cnf.full_adder(a, b, c)
    all_inputs = sequence([[a, ~a], [b, ~b], [c, ~c]])
    test_constraints = [cnf & t[0] & t[1] & t[2] for t in all_inputs]
    return [cnf.as_unigen_string() for cnf in test_constraints]


def soln_full_adder() -> List[str]:
    cnf = CNF()
    a, b, c = cnf.get_n_fresh(3)
    all_inputs = sequence([[a, ~a], [b, ~b], [c, ~c]])
    int_inputs = [[int(v) for v in sl] for sl in all_inputs]
    return ["s SATISFIABLE\nv " + ' '.join(map(str, compute_soln_full_adder(t, 4, 5))) + " 0\n"
            for t in int_inputs]


def compute_soln_full_adder(incoming: List[int], c_index: int, s_index: int) -> List[int]:
    total = sum(map(lambda x: 0 if x < 0 else 1, incoming))
    c = c_index if total > 1 else (-c_index)
    s = s_index if total & 1 else (-s_index)
    return incoming + [c] + [s]


def ripple_carry_dimacs(num_digits: int) -> List[str]:
    cnf = CNF()
    variables = cnf.get_n_fresh(num_digits)
    # TODO: Verify that it's correct to use variables that are not part of the
    #       CNF formula for the second argument to ripple_carry.
    cnf.ripple_carry(variables,
                     [Var(n) for n in range(num_digits + 1, (num_digits * 2) + 1)])
    all_inputs = permute_complements(variables)
    result_cnfs = [CNF([[x] for x in xs]) + cnf for xs in all_inputs]
    return [result_cnf.as_dimacs_string() for result_cnf in result_cnfs]


def pop_count_dimacs(num_digits: int) -> List[str]:
    cnf = CNF()
    variables = cnf.get_n_fresh(num_digits)
    cnf.pop_count(variables, 0)
    all_inputs = permute_complements(variables)
    result_cnfs = [CNF([[x] for x in xs]) + cnf for xs in all_inputs]
    return [result_cnf.as_dimacs_string() for result_cnf in result_cnfs]


def permute_complements(xs: List[T],
                        to_position: Optional[int] = None,
                        complement_func: Callable[[T], T] = invert) -> List[List[T]]:
    """Given a list of elements, computes a list of each possible permutation
    of that list with complements. That is, given a list like:

        [1, 2, 3]

    permute_complements will return:

        [[1, 2, 3], [-1, 2, 3], [1, -2, 3], [-1, -2, 3],
         [1, 2, -3], [-1, 2, -3], [1, -2, -3], [-1, -2, -3]]
    """
    if not xs:
        return []
    if to_position is None:
        return permute_complements(xs, len(xs) - 1, complement_func)
    elif to_position == 0:
        return [[xs[0]], [complement_func(xs[0])]]
    else:
        results: List[List[T]] = []
        permuted_xss = permute_complements(xs, to_position - 1, complement_func)
        for permuted_xs in permuted_xss:
            results.append([x for x in permuted_xs] + [xs[to_position]])
        for permuted_xs in permuted_xss:
            results.append([x for x in permuted_xs] + [complement_func(xs[to_position])])
        return results


def assert_k_of_n_dimacs(num_digits: int, k: int) -> str:
    cnf = CNF()
    variables = cnf.get_n_fresh(num_digits)
    cnf.assert_k_of_n(k, variables)
    return cnf.as_dimacs_string()


def assert_all_k_of_n_dimacs(num_digits: int) -> List[str]:
    return [assert_k_of_n_dimacs(num_digits, k) for k in range(num_digits + 1)]


def pop_count_k_less_than_n_dimacs(num_digits: int, k: int) -> str:
    cnf = CNF()
    variables = cnf.get_n_fresh(num_digits)
    cnf.assert_k_less_than_n(k, variables)
    return cnf.as_dimacs_string()


def pop_count_all_k_less_than_n_dimacs(num_digits: int) -> List[str]:
    return [pop_count_k_less_than_n_dimacs(num_digits, k) for k in range(num_digits + 1)]



"""This module provides simple Docker command-line functionality. It is
generally discouraged to use this functionality, as we prefer using the bundled
executables from `sweetpea-org/unigen-exe
<https://github.com/sweetpea-org/unigen-exe>`_. However, these executables may
lag behind their respective repositories, so we provide the Docker
functionality to stay on the cutting edge.
"""


from subprocess import CompletedProcess, run
from typing import List, Optional

from .return_code import ReturnCodeEnum


#: Whether Docker should be used by default.
DEFAULT_DOCKER_MODE_ON = False


class DockerRunReturnCode(ReturnCodeEnum):
    """Invocation of the ``docker run`` command can result in one of the
    following special return codes. Any other return code is the result of
    invoking the indicated command in the container.

    Based on `this reference information
    <https://docs.docker.com/engine/reference/run/#exit-status>`_.
    """

    #: There was an error executing the Docker daemon itself.
    DockerDaemonError                    = 125
    #: The command produced an error in the container.
    ContainedCommandCannotBeInvokedError = 126
    #: The command specified could not be found in the container.
    ContainedCommandCannotBeFoundError   = 127


class DockerRunError(Exception):
    """An error raised when something fails while invoking Docker."""

    def __init__(self, returncode: DockerRunReturnCode, stderr: str):
        super().__init__(f"Error running Docker: {returncode.name}\n    stderr output captured below:\n\n{stderr}")


def docker_run(container: str,
               args: Optional[List[str]] = None,
               input_bytes: Optional[bytes] = None) -> CompletedProcess:
    """Runs a Docker container, with the optional arguments and input if
    provided.

    If the execution produces an error, a :class:`DockerRunError` will be
    raised.
    """
    if args is None:
        args = []
    command = ['docker', 'run', *args, container]
    # NOTE: flake8 doesn't seem to handle the calls to `run` correctly, but
    #       mypy reports everything is fine here so we `noqa` to prevent flake8
    #       complaining about what it doesn't understand.
    result = run(command, capture_output=True, input=input_bytes)  # noqa
    if DockerRunReturnCode.has_value(result.returncode):
        code = DockerRunReturnCode(result.returncode)
        raise DockerRunError(code, result.stderr.decode())
    return result



"""This module provides functionality for interacting with CNF formulas.

SweetPea offers three mechanisms for CNF interaction:

#. Sampling solutions from a CNF formula uniformly via :func:`sample_uniform`.
#. Sampling solutions from a CNF formula *non*-uniformly via
   :func:`sample_non_uniform`.
#. Determining whether a CNF formula is satisfiable via :func:`is_satisfiable`.
"""


from .is_satisfiable import cnf_is_satisfiable
from .sample_non_uniform import sample_non_uniform, sample_non_uniform_from_specification
from .sample_uniform import sample_uniform
from .utility import AssertionType, GenerationRequest, SampleType, ProblemSpecification, Solution, combine_cnf_with_requests




class BeforeStart:
    def __init__(self, ready_at: int) -> None:
        self.ready_at = ready_at # 0-based

    def __str__(self) -> str:
        return f"BeforeStart<{self.ready_at}>"

    def __repr__(self) -> str:
        return self.__str__()



import operator as op
import pytest

from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition, Window


# Common variables for stroop.
color = Factor("color", ["red", "blue"])
text  = Factor("text",  ["red", "blue"])

con_level  = DerivedLevel("con", WithinTrial(op.eq, [color, text]))
inc_level  = DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
con_factor = Factor("congruent?", [con_level, inc_level])

color_repeats_factor = Factor("color repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[1], [color]))
])

text_repeats_factor = Factor("text repeats?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[1], [text])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[1], [text]))
])

congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])



import pytest


def pytest_addoption(parser):
    parser.addoption('--run-slow', action='store_true', default=False,
                     help="run slow tests")


def pytest_configure(config):
    config.addinivalue_line('markers', 'slow: mark test as slow to run')


def pytest_collection_modifyitems(config, items):
    if config.getoption('--run-slow'):
        # Do not skip slow tests.
        return
    skip_slow = pytest.mark.skip(reason="need --run-slow option to run")
    for item in items:
        if 'slow' in item.keywords:
            item.add_marker(skip_slow)



import pytest

from itertools import permutations

from sweetpea import * 
from sweetpea._internal.server import build_cnf
from acceptance import path_to_cnf_files, reset_expected_solutions

# Basic setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

# congruent 'bookend' Factor. (color and text in first and last trial are congruent)
congruent_bookend = Factor("congruent bookend?", [
    DerivedLevel("yes", Window(lambda color, text: color == text, [color, text], 1, 3)),
    DerivedLevel("no",  Window(lambda color, text: color != text, [color, text], 1, 3))
])


@pytest.mark.parametrize('design', permutations([color, text, congruent_bookend]))
def test_correct_solution_count_when_unconstrained(design):
    crossing = [color, text]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 24


@pytest.mark.parametrize('design', permutations([color, text, congruent_bookend]))
def test_correct_solution_count_when_bookends_must_be_congruent(design):
    crossing = [color, text]

    # Require both bookends to be congruent.
    constraints = [Exclude(congruent_bookend["no"])]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 4


@pytest.mark.parametrize('design', permutations([color, text, congruent_bookend]))
def test_correct_solution_count_when_bookends_must_not_be_congruent(design):
    crossing = [color, text]

    # Require both bookends to not be congruent.
    constraints = [Exclude(congruent_bookend["yes"])]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 4


@pytest.mark.parametrize('design', permutations([color, text, congruent_bookend]))
def test_correct_solution_count_when_bookends_must_not_match_each_other(design):
    crossing = [color, text]

    # Require both bookends to be incongruent with each other.
    constraints = [AtMostKInARow(1, congruent_bookend)]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 16

def test_correct_solution_count_when_bookends_must_not_match_each_other_cnf(design=[color, text, congruent_bookend]):
    crossing = [color, text]

    # Require both bookends to be incongruent with each other.
    constraints = [AtMostKInARow(1, congruent_bookend)]

    block  = CrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_when_bookends_must_not_match_each_other.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_when_bookends_must_not_match_each_other.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()



import pytest
import operator as op

from itertools import permutations

from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition
from sweetpea._internal.constraint import AtMostKInARow
from sweetpea._internal.encoding_diagram import print_encoding_diagram
from sweetpea import CrossBlock, synthesize_trials, print_experiments, IterateGen
from sweetpea._internal.server import build_cnf
from acceptance import path_to_cnf_files, reset_expected_solutions


direction = Factor("direction", ["up", "down"])

color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

congruent_factor = Factor("congruent?", [
    DerivedLevel("con", WithinTrial(op.eq, [color, text])),
    DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
])

repeated_color_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])


@pytest.mark.parametrize('design', [[direction, color, text, congruent_factor],
                                    [congruent_factor, direction, color, text],
                                    [direction, congruent_factor, color, text]])
def test_correct_solution_count_when_unconstrained(design):
    crossing = [direction, congruent_factor]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 500, IterateGen)

    assert len(experiments) == 384


@pytest.mark.parametrize('design', permutations([direction, color, text, congruent_factor]))
def test_correct_solution_count_when_constrained(design):
    crossing = [direction, congruent_factor]
    constraints = [AtMostKInARow(1, congruent_factor)]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 500, IterateGen)

    assert len(experiments) == 128


@pytest.mark.parametrize('design', permutations([direction, color, repeated_color_factor]))
def test_correct_solution_count_when_transition_in_crossing_and_unconstrained(design):
    crossing = [direction, repeated_color_factor]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 96


@pytest.mark.parametrize('design', permutations([direction, color, repeated_color_factor]))
def test_correct_solution_count_when_transition_in_crossing_and_constrained(design):
    crossing = [direction, repeated_color_factor]
    constraints = [AtMostKInARow(1, (color, "red"))]

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 32


def test_correct_solution_count_when_transition_in_crossing_and_constrained_cnf(design=[direction, color, repeated_color_factor]):
    crossing = [direction, repeated_color_factor]
    constraints = [AtMostKInARow(1, (color, "red"))]

    block  = CrossBlock(design, crossing, constraints)
    cnf = build_cnf(block)

    if reset_expected_solutions:
        with open(path_to_cnf_files+'/test_correct_solution_count_when_transition_in_crossing_and_constrained.cnf', 'w') as f:
            f.write(cnf.as_unigen_string())
    with open(path_to_cnf_files+'/test_correct_solution_count_when_transition_in_crossing_and_constrained.cnf', 'r') as f:
        old_cnf = f.read()

    assert old_cnf == cnf.as_unigen_string()


@pytest.mark.parametrize('design', permutations([color, repeated_color_factor]))
def test_correct_solution_count_when_crossing_with_derived_transition(design):
    crossing = [color, repeated_color_factor]
    constraints = []

    block  = CrossBlock(design, crossing, constraints)
    experiments  = synthesize_trials(block, 100, IterateGen)

    assert len(experiments) == 4



import networkx as nx
import operator as op

from sweetpea._internal.primitive import Factor, DerivedLevel, WithinTrial, Transition
from sweetpea._internal.design_graph import DesignGraph


# Setup
color_list = ["red", "blue"]
color = Factor("color", color_list)
text  = Factor("text",  color_list)

# Congruent factor
congruent = Factor("congruent?", [
    DerivedLevel("con", WithinTrial(op.eq, [color, text])),
    DerivedLevel("inc", WithinTrial(op.ne, [color, text]))
])

# Repeated color factor
repeated_color_factor = Factor("repeated color?", [
    DerivedLevel("yes", Transition(lambda colors: colors[0] == colors[-1], [color])),
    DerivedLevel("no",  Transition(lambda colors: colors[0] != colors[-1], [color]))
])

# Repeated text factor
repeated_text_factor = Factor("repeated text?", [
    DerivedLevel("yes", Transition(lambda texts: texts[0] == texts[-1], [text])),
    DerivedLevel("no",  Transition(lambda texts: texts[0] != texts[-1], [text]))
])


def test_basic_graph():
    dg = DesignGraph([color, text, congruent])

    assert len(dg.graph.nodes()) == 3
    assert dg.graph.has_node("color")
    assert dg.graph.has_node("text")
    assert dg.graph.has_node("congruent?")

    assert len(dg.graph.edges()) == 2
    assert dg.graph.has_edge("congruent?", "color")
    assert dg.graph.has_edge("congruent?", "text")

    # Graph is directed, so these should not exist.
    assert not dg.graph.has_edge("color", "congruent?")
    assert not dg.graph.has_edge("text", "congruent?")


def test_graph_paths():
    dg = DesignGraph([color, text, congruent, repeated_color_factor, repeated_text_factor])

    assert nx.has_path(dg.graph, "congruent?", "color")
    assert nx.has_path(dg.graph, "congruent?", "text")

    assert nx.has_path(dg.graph, "repeated color?", "color")
    assert nx.has_path(dg.graph, "repeated text?", "text")

    assert not nx.has_path(dg.graph, "repeated color?", "text")
    assert not nx.has_path(dg.graph, "repeated text?", "color")



import glob
import os
import pytest

from sweetpea._internal.logic import And, cnf_to_json
from sweetpea._internal.server import build_cnf
from sweetpea._internal.core import cnf_is_satisfiable, CNF
from sweetpea._internal.sampling_strategy.random import UCSolutionEnumerator


path_to_test_files = os.path.dirname(os.path.abspath(__file__)) + "/../sweetpea/tests/sampling_strategies/uc-counting-tests/*.py"

@pytest.mark.slow
@pytest.mark.parametrize('filename', glob.glob(path_to_test_files))
def test_uniform_combinatoric_is_always_valid(filename):

    failures = []
    contents = None
    with open(filename, 'r') as f:
        contents = f.read()
        exec(contents, globals(), locals())

    if 'block' not in vars():
        pytest.fail("File did not produce a variable named 'block', aborting. file={}".format(filename))


    # Build the CNF for this block
    build_cnf_result = build_cnf(vars()['block'])

    # Build the sampler
    enumerator = UCSolutionEnumerator(vars()['block'])

    # Generate some samples, make sure they're all SAT.
    sample_count = min(enumerator.solution_count(), 200)
    print("Checking that UC samples are SAT for {}, sample count={}".format(filename, sample_count))
    for s in range(sample_count):
        sample = enumerator.generate_solution_variables()
        if not cnf_is_satisfiable(build_cnf_result + CNF(cnf_to_json([And(sample)]))):
            failures.append("Found UNSAT solution! Solution={} File={}".format(sample, filename))

    if failures:
        pytest.fail('{} failures occurred in SAT checks for UC sampler: {}'.format(len(failures), failures))



import pytest

from sweetpea import *
from acceptance import assert_atmostkinarow, shuffled_design_sample

# Simple Factors
color  = Factor("color",  ["red", "blue"])
motion = Factor("motion", ["up", "down"])
task   = Factor("task",   ["color", "motion"])

# Response Definition
def response_left(task, color, motion):
    return (task == "color"  and color  == "red") or \
           (task == "motion" and motion == "up")

def response_right(task, color, motion):
    return not response_left(task, color, motion)

response = Factor("response", [
    DerivedLevel("left",  WithinTrial(response_left,  [task, color, motion])),
    DerivedLevel("right", WithinTrial(response_right, [task, color, motion]))
])

# Congruency Definition
def color_motion_congruent(color, motion):
    return ((color == "red") and (motion == "up")) or \
           ((color == "blue") and (motion == "down"))

def color_motion_incongruent(color, motion):
    return not color_motion_congruent(color, motion)

congruency = Factor("congruency", [
    DerivedLevel("con", WithinTrial(color_motion_congruent,   [color, motion])),
    DerivedLevel("inc", WithinTrial(color_motion_incongruent, [color, motion]))
])

# Task Transition
task_transition = Factor("task Transition", [
    DerivedLevel("repeat", Transition(lambda tasks: tasks[0] == tasks[-1], [task])),
    DerivedLevel("switch", Transition(lambda tasks: tasks[0] != tasks[-1], [task]))
])

# Response Transition
response_transition = Factor("response Transition", [
    DerivedLevel("repeat", Transition(lambda responses: responses[0] == responses[-1], [response])),
    DerivedLevel("switch", Transition(lambda responses: responses[0] != responses[-1], [response]))
])


@pytest.mark.parametrize('design', shuffled_design_sample([color, motion, task, response, congruency, task_transition, response_transition], 6))
def test_that_design_is_correctly_constrained(design):
    crossing = [color, motion, task]

    k = 2
    constraints = [
        AtMostKInARow(k, task_transition),
        AtMostKInARow(k, response_transition)
    ]

    block = CrossBlock(design, crossing, constraints)
    experiments = synthesize_trials(block, 100)

    assert len(experiments) == 100, "Design: %s" % str(list(map(lambda f: f.factor_name, design)))
    for c in constraints:
        assert_atmostkinarow(c, experiments)



import pytest

from sweetpea import *

number = Factor("number", [1.0, 2.0, 3.0])
number2 = Factor("number", [Level(1.0, weight=2), 2.0, 3.0])
color = Factor("color", ["red", "green", "blue"])

def check_mutually_exclusive(ex1, ex2, strategy):
    trials1 = synthesize_trials(ex1, 10, strategy)
    trials2 = synthesize_trials(ex2, 10, strategy)

    assert len(trials1) > 0
    assert len(trials2) > 0

    for trials in trials1:
        assert sample_mismatch_experiment(ex2, trials)
    for trials in trials2:
        assert sample_mismatch_experiment(ex1, trials)

@pytest.mark.parametrize('strategy', [IterateSATGen, RandomGen])
def test_mutually_exclusive_experiments(strategy):
    # Try pairs of experiments where a solution to one is
    # never a solution to the other

    check_mutually_exclusive(CrossBlock([number], [number], []),
                             CrossBlock([number2], [number2], []),
                             strategy)

    check_mutually_exclusive(CrossBlock([number], [number], [Pin(1, number[1.0])]),
                             CrossBlock([number], [number], [Pin(1, number[2.0])]),
                             strategy)

    check_mutually_exclusive(CrossBlock([number], [number], [Pin(1, number[1.0])]),
                             CrossBlock([number], [number], [Pin(-1, number[1.0])]),
                             strategy)

    check_mutually_exclusive(CrossBlock([number], [number], [Exclude(number[2.0])], False),
                             CrossBlock([number], [number], []),
                             strategy)

    check_mutually_exclusive(CrossBlock([number2], [number2], [AtMostKInARow(1, number2[1.0])]),
                             CrossBlock([number2], [number2], [AtLeastKInARow(2, number2[1.0])]),
                             strategy)

    check_mutually_exclusive(CrossBlock([number, color], [number, color], [AtMostKInARow(1, number[1.0])]),
                             CrossBlock([number, color], [number, color], [AtLeastKInARow(2, number[1.0])]),
                             strategy)

    check_mutually_exclusive(CrossBlock([number, color], [number, color], [MinimumTrials(18)]),
                             CrossBlock([number, color], [number, color], []),
                             strategy)

def test_mutually_exclusive_experiments_SAT():
    check_mutually_exclusive(CrossBlock([number, color], [number, color], [MinimumTrials(18),
                                                                           # implies that all "red"s are together,
                                                                           # would have to be in the middle to
                                                                           # satisfy repeat construction
                                                                           AtLeastKInARow(4, color["red"]),
                                                                           # ditto (so can't both happen)
                                                                           AtLeastKInARow(4, number[2.0])]),
                             Repeat(CrossBlock([number, color], [number, color], []),
                                    [MinimumTrials(18)]),
                             IterateSATGen)



"""This module provides functionality for collecting metrics about a specific
block in an experimental design.
"""


from math import factorial
from typing import Dict

from sweetpea._internal.blocks import Block
from sweetpea._internal.main import __generate_cnf


def collect_design_metrics(block: Block) -> Dict:
    """Given a block, this function will collect various metrics pertaining to
    the block and return them in a dictionary.
    """
    backend_request = block.build_backend_request()
    dimacs_header = __generate_cnf(block).split('\n')[0].split(' ')

    return {
        'full_factor_count': len(block.design),
        'non_implied_factor_count': len(block.act_design),
        'crossing_factor_count': len(block.crossings),
        'constraint_count': len(block.constraints),

        'block_length': block.trials_per_sample(),
        'block_length_factorial': factorial(block.trials_per_sample()),

        'low_level_request_count': len(backend_request.ll_requests),
        'cnf_total_variables': int(dimacs_header[2]),
        'cnf_total_clauses': int(dimacs_header[3])
    }
